from gym.envs.registration import register

from .session.session import Session

# register test envs with gym
register(
        id = "Obstacle-v1",
        entry_point = "graph_rl.envs:ObstacleEnv"
        )




import numpy as np

import gym
from gym import spaces
from gym.utils import seeding

from .graphics_utils import ArrowConfig, get_default_subgoal_colors

class ObstacleEnv(gym.GoalEnv):  
    metadata = {'render.modes': ['human']}   

    def __init__(self, obstacle_radius=0.5, stage_dimension=6., 
            agent_speed=5e-2, max_episode_length=300, subgoal_radius=0.033):
        super().__init__()

        self.stage_dimension = stage_dimension
        self.obstacle_radius = obstacle_radius
        self.agent_speed = agent_speed
        self.max_episode_length = max_episode_length

        desired_goal_space = spaces.Box(
                low = -1., 
                high = 1., 
                shape = (2,),
                dtype = np.float32)
        achieved_goal_space = desired_goal_space
        obs_space = desired_goal_space

        self.observation_space = spaces.Dict({
            "observation": obs_space,
            "desired_goal": desired_goal_space,
            "achieved_goal": achieved_goal_space
            })

        self.action_space = spaces.Box(
                low = -1., 
                high = 1., 
                shape = (2,),
                dtype = np.float32)

        self.window = None

        self.window_width = 800
        self.window_height = 800
        self.background_color = (1.0, 1.0, 1.0, 1.0)

        self.obstacle_color = (0.4, 0.4, 0.4)

        self.agent_position = np.array((0., -1.0))
        self.agent_radius = 0.1
        self.agent_color = (0.0, 0.0, 0.0)

        self._draw_goal()
        self.goal_radius = 0.1
        self.goal_color = (0.0, 0.0, 0.0)

        self.current_step = 0

        self._subgoals = []
        self._timed_subgoals = []
        self._tolerances = []
        self._subgoal_colors = get_default_subgoal_colors()
        self.subgoal_radius = float(subgoal_radius)*self.stage_dimension*0.5

        self.function_grid = None 
        self.color_low = np.array((0., 0., 1.))
        self.color_high = np.array((1., 0., 0.))
        self.value_low = -10.
        self.value_high = 0.

    def _draw_goal(self):
        while True:
            candidate = np.random.normal((0., 1.), 0.3, size = (2,))
            if np.linalg.norm(candidate) > self.agent_radius + self.obstacle_radius:
                self.goal = candidate
                break

    def update_function_grid(self, values, low, high):
        self.function_grid = values
        self.value_low = low
        self.value_high = high

    def compute_reward(self, achieved_goal, desired_goal, info):
        if np.linalg.norm(achieved_goal - desired_goal) <= \
                self.goal_radius/self.stage_dimension/0.5:
                    return 0.
        else:
            return -1.

    def _get_obs(self):
        obs = { 
            "observation": self.agent_position/self.stage_dimension*2.,
            "desired_goal" : self.goal/self.stage_dimension*2.,
            "achieved_goal": self.agent_position/self.stage_dimension*2.
            }
        return obs

    @classmethod
    def map_to_env_goal(self, partial_obs):
        return partial_obs

    def step(self, action):
        self.agent_position += self.agent_speed*np.array(action)
        self.agent_position = np.clip(self.agent_position,
                -0.5*self.stage_dimension, 0.5*self.stage_dimension)
        distance_to_center = np.linalg.norm(self.agent_position)
        if distance_to_center < self.obstacle_radius + self.agent_radius:
            self.agent_position *= (self.obstacle_radius + self.agent_radius)/distance_to_center
        info = {}
        obs = self._get_obs()
        reward = self.compute_reward(self.agent_position/self.stage_dimension*2., 
                self.goal/self.stage_dimension*2., info)
        self.current_step += 1
        done = reward == 0. or self.current_step >= self.max_episode_length
        return obs, reward, done, info
  
    def reset(self):
        self.agent_position = np.array((0., -1.0))
        self.current_step = 0
        self._draw_goal()
        return self._get_obs()

    def update_subgoals(self, subgoals):
        self._subgoals = [np.array(sg)*self.stage_dimension*0.5 for sg in subgoals]

    def update_timed_subgoals(self, timed_subgoals, tolerances):
        self._timed_subgoals = timed_subgoals
        self._tolerances = tolerances
        for ts in self._timed_subgoals:
            if ts is not None:
                ts.goal = self.stage_dimension*0.5*ts.goal

    def render(self, mode='human', close=False):
        import pyglet
        import pyglet.gl as gl

        from .pyglet_utils import (draw_circle_sector, draw_box, draw_line, draw_vector, draw_vector_with_outline, 
                draw_circular_subgoal)

        if self.window is None:
            self.window = pyglet.window.Window(width = self.window_width,
                                               height = self.window_height,
                                               vsync = True,
                                               resizable = True)
            gl.glClearColor(*self.background_color)

        @self.window.event
        def on_resize(width, height):
            gl.glViewport(0, 0, width, height)
            gl.glMatrixMode(gl.GL_PROJECTION)
            gl.glLoadIdentity()
            gl.glOrtho(-0.5*self.stage_dimension, 
                       0.5*self.stage_dimension, 
                       -0.5*float(height)/width*self.stage_dimension,
                       0.5*float(height)/width*self.stage_dimension, 
                       -1., 
                       1.)
            gl.glMatrixMode(gl.GL_MODELVIEW)
            return pyglet.event.EVENT_HANDLED


        def draw_function_grid(values, a, b, color_low, color_high, value_low, value_high):
            diff = np.array(b) - np.array(a)
            diff[0] /= values.shape[0]
            diff[1] /= values.shape[1]
            v_diff = value_high - value_low
            for j in range(values.shape[1]):
                gl.glBegin(gl.GL_TRIANGLE_STRIP)
                for i in range(values.shape[0]):
                    value = np.clip(values[i, j], value_low, value_high)
                    v = (value - value_low)/v_diff
                    gl.glColor3f(*list(v*color_high + (1. - v)*color_low))
                    gl.glVertex2f(a[0] + diff[0]*i, a[1] + diff[1]*j)

                    value = np.clip(values[i, min(j + 1, values.shape[1] - 1)], value_low, value_high)
                    v = (value - value_low)/v_diff
                    gl.glColor3f(*list(v*color_high + (1. - v)*color_low))
                    gl.glVertex2f(a[0] + diff[0]*i, a[1] + diff[1]*(j + 1))
                gl.glVertex2f(a[0] + diff[0]*values.shape[0], a[1] + diff[1]*j)
                gl.glVertex2f(a[0] + diff[0]*values.shape[0], a[1] + diff[1]*(j + 1))
                gl.glEnd()


        def draw_timed_circular_subgoal(position, delta_t_ach, delta_t_comm, radius, color):
            draw_circular_subgoal(position, None, radius, color, None)
            # desired time until achievement
            draw_box(position + (0., radius + 0.05), delta_t_ach/100., 0.03, 0., color)
            # remaining commitment time
            draw_box(position + (0., radius + 0.02), delta_t_comm/100., 0.03, 0., (0., 0., 0.))


        self.window.clear()
        self.window.switch_to()
        self.window.dispatch_events()

        gl.glLoadIdentity()

        n_triangles = 32

        if self.function_grid is not None:
            draw_function_grid(self.function_grid, -0.5*self.stage_dimension*np.ones(2), 
                    0.5*self.stage_dimension*np.ones(2), self.color_low, self.color_high, 
                    self.value_low, self.value_high)

        # obstacle
        draw_circle_sector([0., 0.], 
                0., 
                self.obstacle_radius,
                n_triangles, 
                self.obstacle_color,
                n_triangles)

        # subgoals
        for subgoal, color in zip(self._subgoals, self._subgoal_colors):
            draw_circle_sector(subgoal, 
                    0., 
                    self.subgoal_radius,
                    n_triangles, 
                    (0., 0., 0.),
                    n_triangles)
            draw_circle_sector(subgoal, 
                    0., 
                    0.8*self.subgoal_radius,
                    n_triangles, 
                    color,
                    n_triangles)

        # timed subgoals
        for ts, color, tol in zip(self._timed_subgoals, self._subgoal_colors, self._tolerances):
            if ts is not None:
                r = tol if tol is not None else self.subgoal_radius
                draw_timed_circular_subgoal(ts.goal, ts.delta_t_ach, 
                        ts.delta_t_comm, r, color)

        # goal
        draw_circle_sector(self.goal, 
                0., 
                self.goal_radius,
                n_triangles, 
                self.goal_color,
                n_triangles)
        draw_circle_sector(self.goal, 
                0., 
                self.goal_radius*0.8,
                n_triangles, 
                self.background_color[:3],
                n_triangles)

        # agent 
        draw_circle_sector(self.agent_position, 
                0., 
                self.agent_radius,
                n_triangles, 
                self.agent_color,
                n_triangles)


        self.window.flip()


class ObstacleEnvHAC(ObstacleEnv):
    """Version of obstalce environment that uses interface
    required by implementation of HAC by Andrew Levy."""

    def __init__(self):
        super().__init__()

        self.name = "ObstacleEnvHAC"

        self.max_actions = int(self.max_episode_length)
        self.visualize = False
        self.visualize_every_nth_episode = 10
        self.episode_counter = 0


        # Projection functions 
        self.project_state_to_end_goal = lambda s, s2: s
        self.project_state_to_subgoal = lambda s, s2: s

        # variables needed for HAC implementation
        self.state_dim = self.observation_space["observation"].low.shape[0]
        self.action_dim = self.action_space.low.shape[0] # low-level action dim
        self.action_bounds = np.ones(2) # low-level action bounds
        self.action_offset = np.zeros((len(self.action_bounds))) # Assumes symmetric low-level action ranges
        self.end_goal_dim = self.observation_space["desired_goal"].low.shape[0]
        self.subgoal_dim = 2
        self.subgoal_bounds = np.array([
            [-1., 1.],
            [-1., 1.],
            ])
        self.subgoal_bounds_symmetric = np.array(
                [1., 1.])
        self.subgoal_bounds_offset = np.zeros((2))

        # End goal/subgoal thresholds
        self.subgoal_thresholds = np.ones(2)*(self.goal_radius/self.stage_dimension*2.)
        self.end_goal_thresholds = np.ones(2)*(self.goal_radius/self.stage_dimension*2.)


    def get_state(self):
        return self._get_obs()["observation"]

    def reset_sim(self):
        # save old goal
        old_goal = self.goal

        # reset (this also overwrites self._desired_goal)
        self.reset()

        # restore old self._desired_goal (in order to not mess up HAC implementation)
        # (Note: If this step is omitted, the goal in the HAC implementation and in 
        # the environment diverge and rendering and reward calculation in the environment
        # are off.)
        self.goal = old_goal

        self.episode_counter += 1

        # Return state
        return self.get_state()

    def render(self):
        super().render(self)

    def execute_action(self, action):
        obs, reward, done, info = self.step(action)
        self.done = done

        if self.visualize:
            if self.episode_counter % self.visualize_every_nth_episode == 0:
                self.render()

        # call the current state "sim" in order to trick the HAC implementation which 
        # expects an underlying Mujoco simulation
        self.sim = self.get_state()

        return self.sim

    def display_end_goal(self, end_goal):
        pass

    def get_next_goal(self, test):
        self._draw_goal()
        return self.goal/self.stage_dimension*2.0

    def display_subgoals(self, subgoals):
        self.update_subgoals(subgoals)

    def set_visualization(self, visualize, visualize_every_nth_episode):
        self.visualize = visualize
        self.visualize_every_nth_episode = visualize_every_nth_episode



from . import AuxReward

class GoalTolReward(AuxReward):
    """Auxiliary reward based on chosen goal_tol."""

    def __init__(self, func):
        self._f = func

    def __call__(self, obs, action, env_reward):
        return self._f(action["goal_tol"])



import torch
from torch import nn

class DDPGNetwork(nn.Module):
    """Outputs action and state (None) as required by Tianshou implementation of DDPG.
    
    The output of input_processing is squashed using tanh."""
    def __init__(self, input_processing, max_action, n_hidden, output_dim, device):
        """
        Args: 
            input_processing: pytorch module
        """
        super().__init__()
        self.device = device
        self.input_processing = input_processing
        self.max_action = torch.tensor(max_action)
        self.mu_fn = nn.Linear(n_hidden, output_dim)

    def forward(self, s, **kwargs):
        if not isinstance(s, torch.Tensor):
            s = torch.tensor(s, device = self.device, dtype = torch.float)
        batch_size = s.shape[0]
        s_flat = s.view(batch_size, -1)
        orig_output = self.input_processing(s_flat)
        mu = self.mu_fn(orig_output)
        action = self.max_action*torch.tanh(mu)
        return action, None



import numpy as np
from gym import spaces as gym_spaces

from .space import Space

class BoxSpace(Space):

    def __init__(self, low, high, dtype = np.float32):
        super().__init__()

        assert np.shape(low) == np.shape(high)

        self.low = np.array(low)
        self.high = np.array(high)
        self.dtype = dtype
        self._difference = self.high - self.low
        self.n = np.shape(self.low)[0]

        self.bounded_below = -np.inf < self.low
        self.bounded_above = self.high < np.inf

    def contains(self, value):
        return np.all(np.logical_and(self.low < value, value < self.high))

    def get_flat_space(self, indices = None):
        if indices is None:
            return self
        else:
            subspace = BoxSpace(
                    low = self.low[indices],
                    high = self.high[indices], 
                    dtype = self.dtype
                    )
            return subspace

    def get_flat_dim(self):
        """Get dimension of flat space associated to this space."""
        return self.n

    def flatten_value(self, value):
        return value

    def unflatten_value(self, flat_value):
        """Maps a vector from the space returned by flatten_value back to this space."""
        return flat_value

    def sample(self):
        res = np.empty(self.low.shape)
        for a in [False, True]:
            for b in [False, True]:
                applies = (b == self.bounded_below) & (a == self.bounded_above)
                # bounded
                if b == True and a == True:
                    res[applies] = self._difference[applies]*np.random.random(size = np.sum(applies)) \
                            + self.low[applies]
                # bounded from below
                if b == True and a == False:
                    res[applies] = np.random.exponential(size = np.sum(applies)) + self.low[applies]
                # bounded from above
                if b == False and a == True:
                    res[applies] = -np.random.exponential(size = np.sum(applies)) + self.high[applies]
                # unbounded
                if b == False and a == False:
                    res[applies] = np.random.normal(size = np.sum(applies))
        return res

    def get_gym_space(self):
        return gym_spaces.Box(self.low, self.high, dtype = self.dtype)

    def __eq__(self, other):
        if not isinstance(other, BoxSpace):
            return False
        return np.all(np.equal(self.low, other.low)) \
                and np.all(np.equal(self.high, other.high)) \
                and self.dtype == other.dtype

    @staticmethod
    def concatenate(*args):
        for arg in args[1:]:
            assert arg.dtype == args[0].dtype
        low = np.concatenate([arg.low for arg in args])
        high = np.concatenate([arg.high for arg in args])
        return BoxSpace(low, high, dtype = args[0].dtype)

    @staticmethod
    def get_from_gym_space(gym_space):
        return BoxSpace(low = gym_space.low, high = gym_space.high, 
                dtype = gym_space.dtype)

    def __str__(self):
        return "BoxSpace(low = {}; high = {}; dtype = {})".format(self.low, self.high, self.dtype)



from abc import ABC, abstractmethod

class Policy(ABC):

    def __init__(self, name, observation_space, action_space):
        super().__init__()
        self.name = name
        self.observation_space = observation_space
        self.action_space = action_space

    @abstractmethod
    def __call__(self, observation, algo_info, testing = False):
        pass



from typing import Any, Tuple, Union, Optional

import numpy as np

from tianshou.data import ReplayBuffer
from tianshou.data.batch import Batch

class SegmentedReplayBuffer:

    def __init__(self, sizes: dict, ratios = None, **kwargs) -> None:
        self._buffers = {name: ReplayBuffer(size) for name, size in sizes.items()}
        self._len = sum(sizes.values())
        if ratios is not None:
            assert sizes.keys() == ratios.keys()
        self._ratios = ratios

    def add(self,
            name, 
            obs: Union[dict, np.ndarray],
            act: Union[np.ndarray, float],
            rew: float,
            done: bool,
            obs_next: Optional[Union[dict, np.ndarray]] = None,
            info: dict = {},
            policy: Optional[Union[dict, Batch]] = {},
            **kwargs) -> None:
        """Add a batch of data into replay buffer."""
        self._buffers[name].add(obs, act, rew, done, obs_next, info, policy, **kwargs)

    def sample_and_process(self, batch_size: int, process_fn: callable) -> Batch:
        """Get a random sample from buffer with size equal to batch_size.

        :return: Sample data.
        """
        if self._ratios is None:
            n_total = float(self.__len__())
            batch_sizes = {key: int(batch_size*len(buff)/n_total) for key, buff 
                    in self._buffers.items()}
        else:
            batch_sizes = {key: int(batch_size*ratio) for key, ratio in self._ratios.items()}
        # make sure the overall batch size is as desired
        batch_sizes[list(self._buffers.keys())[0]] += batch_size - sum(batch_sizes.values())

        batch = Batch()
        for name, bs in batch_sizes.items():
            sub_batch, sub_indices = self._buffers[name].sample(bs)
            sub_batch = process_fn(sub_batch, self._buffers[name], sub_indices)
            batch.cat(sub_batch)

        return batch

    def __len__(self) -> int:
        return sum(len(b) for b in self._buffers.values())







import numpy as np

from . import AuxReward

class DeltaTAchReward(AuxReward):
    """Auxiliary reward proportional to minus the time until achievement.
    
    The time until achievement (delta_t_ach) refers to the quantity 
    used in the HiTS algorithm. This auxiliary reward can therefore 
    be applied to a parent node of a HiTS node. 
    Note that the time until achievement is normalized to the inverval, 
    [0, 1]."""

    def __init__(self, weight):
        """Weight refers to the weight of this reward in the overall reward."""

        self._w = weight

    def __call__(self, obs, action, env_reward):
        delta_t_ach_norm = 0.5*(action["delta_t_ach"][0] + 1.0)
        return -self._w*delta_t_ach_norm



from .mlp_network import MLPNetwork
from .concat_network import ConcatNetwork
from .sac_network import SACNetwork
from .ddpg_network import DDPGNetwork
from .squash_network import SquashNetwork
from .negative_network import NegativeNetwork




class SessInfo:
    def __init__(self, ep_step = None, total_step = None, learn = True, 
            testing = False):
        self.ep_step = ep_step
        self.total_step = total_step
        self.learn = learn
        self.testing = testing



import numpy as np
from gym import spaces as gym_spaces

from .space import Space

class BinarySpace(Space):
    def __init__(self):
        super().__init__()
        self.dtype = bool

    def contains(self, value):
        if not isinstace(value, np.bool_) or isinstance(value, bool):
            return False
        else:
            return True

    def get_flat_space(self):
        return self

    def flatten_value(self, value):
        return np.array([value])

    def get_flat_dim(self):
        """Get dimension of flat space associated to this space."""
        return 1

    def unflatten_value(self, flat_value):
        """Maps a vector from the space returned by flatten_value back to this space."""
        return flat_value[0]

    def sample(self):
        res = np.random.randint(0, 2, dtype = np.int8)
        return bool(res)

    def get_gym_space(self):
        return gym_spaces.MultiBinary(1, dtype = self.int8)

    def __eq__(self, other):
        return isinstance(other, BinarySpace)

    @staticmethod
    def concatenate(*args):
        for arg in args:
            assert isinstance(arg, BinarySpace)
        return MultiBinarySpace(len(args), dtype = args[0].dtype)

    @staticmethod
    def get_from_gym_space(gym_space):
        return BinarySpace()

    def __str__(self):
        return "BinarySpace"



from abc import ABC, abstractmethod

class GlobalAlgorithm(ABC):

    def __init__(self, name):
        super().__init__()
        self.name = name

    @abstractmethod
    def add_experience(self, transition, sess_info):
        pass

    @abstractmethod
    def learn(self, sess_info):
        pass



from copy import copy
import os

from . import Subtask
from ..spaces import DictSpace
from ..data import SubtaskTransition
from ..logging import CSVLogger

class ShortestPathSubtask(Subtask):
    """Subtask is to achieve a subgoal with as few actions as possible."""

    def __init__(self, name, subtask_spec):
        """Args:
            subtask_spec (ShortestPathSubtaskSpec): Specification of the shortest path subtask.
        """
        super(ShortestPathSubtask, self).__init__(name, subtask_spec)

        self.task_spec = subtask_spec

        self._observation_space = DictSpace({
            "partial_observation": subtask_spec.partial_obs_space, 
            "desired_goal": subtask_spec.goal_space
            })

        self._n_actions_taken = 0

        self.logger_test = None
        self.logger_train = None
        self.logfiles = {}

        self.reset()

    @property
    def observation_space(self):
        return self._observation_space

    @property
    def parent_action_space(self):
        return self.task_spec.parent_action_space

    def reset(self):
        # reset number of actions taken at beginning of episode
        self._n_actions_taken = 0
        self._return = 0

    def get_observation(self, env_obs, parent_info, sess_info):
        partial_obs = self.task_spec.map_to_partial_obs(env_obs, parent_info, 
            sess_info.ep_step)
        desired_goal = self.task_spec.get_desired_goal(env_obs, parent_info)

        full_obs = {
                "partial_observation": partial_obs, 
                "desired_goal": desired_goal, 
                }
        return full_obs

    def check_achievement(self, achieved_goal, desired_goal, obs, action, parent_info, env_info):
        if self.task_spec.goal_achievement_criterion(achieved_goal, desired_goal, parent_info):
            reward = 0.
            achieved = True
        else:
            reward = -1.
            achieved = False

        # add auxiliary rewards
        if obs is not None and action is not None:
            reward += self.get_aux_rewards(obs, action, env_info.reward)
            
        return achieved, reward

    def check_interruption(self, env_info, new_subtask_obs, parent_info, sess_info):
        super().check_interruption(env_info, new_subtask_obs, parent_info, sess_info)
        new_env_obs = env_info.new_obs
        desired_goal = self.task_spec.get_desired_goal(new_env_obs, parent_info)
        new_partial_obs = self.task_spec.map_to_partial_obs(new_env_obs, parent_info, 
            sess_info.ep_step)
        achieved_goal = self.task_spec.map_to_goal(new_partial_obs)
        achieved, _ = self.check_achievement(achieved_goal, desired_goal, None, None, parent_info, env_info)
        return achieved

    def evaluate_transition(self, env_obs, env_info, subtask_trans, parent_info, algo_info, sess_info):
        desired_goal = self.task_spec.get_desired_goal(env_obs, parent_info)
        new_partial_obs = self.task_spec.map_to_partial_obs(env_info.new_obs, parent_info, 
                sess_info.ep_step)
        achieved_goal = self.task_spec.map_to_goal(new_partial_obs)
        achieved, reward = self.check_achievement(achieved_goal, desired_goal, 
                subtask_trans.obs, subtask_trans.action,  parent_info, env_info)

        self._n_actions_taken += 1

        # if constant failure return in wanted, add minus remaining number of actions
        if env_info.done and self.task_spec.constant_failure_return and not achieved:
            reward += (self.task_spec.max_n_actions - self._n_actions_taken)*(0.3*-10. - 1.)

        self._return += reward

        # subtask ended if subgoal achieved or running out of actions
        if achieved or (self.task_spec.max_n_actions is not None and 
                self._n_actions_taken >= self.task_spec.max_n_actions):
            n_actions_taken = self._n_actions_taken
            ret = self._return
            self._return = 0
            self._n_actions_taken = 0
            ended = True
        else:
            ended = False



        # NOTE: The boolean ended only indicates whether the subtask ended, not that the goal was reached!
        # Whether the goal was reached is encoded in the key "has_achieved" in info.
        info = {
                "has_achieved": achieved,
                "achieved_generalized_goal": achieved_goal
                }
        feedback = info

        complete_subtask_trans = copy(subtask_trans)
        complete_subtask_trans.reward = reward
        complete_subtask_trans.ended = ended
        complete_subtask_trans.info = info

        # tensorboard logging
        if self._tb_writer is not None:
            if ended == True:
                mode = "test" if sess_info.testing else "train"
                self._tb_writer.add_scalar(f"{self.name}/{mode}/subgoal_achieved", int(achieved), sess_info.total_step)
                self._tb_writer.add_scalar(f"{self.name}/{mode}/n_actions", n_actions_taken, sess_info.total_step)
                self._tb_writer.add_scalar(f"{self.name}/{mode}/return", ret, sess_info.total_step)
        # csv logging
        if ended and self.logger_train is not None and self.logger_test is not None:
            row_dict = {
                    "achieved": int(achieved), 
                    "n_actions": n_actions_taken, 
                    "step": sess_info.total_step, 
                    "time": self.logger_test.time_passed() if sess_info.testing else self.logger_train.time_passed()
                    }
            if sess_info.testing:
                self.logger_test.log(row_dict)
            elif self.logger_train:
                self.logger_train.log(row_dict)

        return complete_subtask_trans, feedback

    def create_logfiles(self, logdir, append = False):
        if logdir is not None:
            logfile_test = os.path.join(logdir, self.name  + "_test.csv")
            logfile_train = os.path.join(logdir, self.name  + "_train.csv")
            self.logger_test = CSVLogger(logfile_test, ("achieved", "n_actions", "step", "time"), append)
            self.logger_train = CSVLogger(logfile_train, ("achieved", "n_actions", "step", "time"), append)
            self.logfiles["test"] = logfile_test
            self.logfiles["train"] = logfile_train

    def get_logfiles(self):
        return self.logfiles



from abc import ABC, abstractmethod

import numpy as np

from ..spaces import BoxSpace, DictSpace
from ..utils import listify, get_obs_space_from_env, get_obs_from_gym
from .subtask_specs import SubtaskSpec

class TimedGoal:
    def __init__(self, goal, delta_t_ach, delta_t_comm = None):
        self.goal = goal
        self.delta_t_ach = delta_t_ach
        self.delta_t_comm = delta_t_comm

    @staticmethod
    def from_dict(d):
        return TimedGoal(d["desired_goal"], d["delta_t_ach"][0], 
                d["delta_t_comm"][0] if "delta_t_comm" in d else d["delta_t_ach"][0])

class TimedGoalSubtaskSpec(SubtaskSpec):
    """Specification of timed goal subtask, i.e. criterion of goal achievement etc."""

    def __init__(self, goal_space, partial_obs_space, delta_t_ach = None,  
            delta_t_comm = None, delta_t_max = None, delta_t_min = 0.):
        super().__init__()
        self._goal_space = goal_space
        self._partial_obs_space = partial_obs_space
        self._delta_t_ach = delta_t_ach
        self._delta_t_comm = delta_t_comm
        self._delta_t_max = delta_t_max
        self._delta_t_min = delta_t_min
        # parent action space contains goal, desired time until achievement and (#TODO) commitment time
        space_dict = {"goal": self.goal_space}
        if delta_t_ach is None:
            # normalize time to achievement to [-1, 1] and take into account
            # delta_t_min when adding it to action space
            a_min = self.convert_time(delta_t_min)
            space_dict["delta_t_ach"] = BoxSpace(low = [a_min], high = [1.])
        self._parent_action_space = DictSpace(space_dict)

    @abstractmethod
    def goal_achievement_criterion(self, achieved_goal, desired_goal, parent_info):
        """Maps achieved_goal, desired_goal and parent_info to True/False depending on whether desired_goal was achieved.
        
        Note that parent_info can contain information about the tolerance for achieving a subgoal."""
        pass

    def unconvert_time(self, t):
        """Map from [-1, 1] interval to time steps."""

        return (t + 1.)/2.*self._delta_t_max

    def convert_time(self, t):
        """Map time to [-1, 1] interval."""

        return t/self._delta_t_max*2. - 1.

    def get_desired_timed_goal(self, env_obs, parent_info):
        """Maps env_obs and parent_info to desired timed goal."""
        # desired goal is equal to parent action, can overwrite for preprocessing
        tg = TimedGoal(
                goal = parent_info.action["goal"],
                delta_t_ach = parent_info.action["delta_t_ach"][0] \
                        if self._delta_t_ach is None else self._delta_t_ach,
                delta_t_comm = parent_info.action["delta_t_ach"][0] \
                        if self._delta_t_comm is None else self._delta_t_comm)
        return tg

    def get_achieved_timed_goal_dict(self, achieved_goal, delta_t_ach, parent_info):
        """delta_t_ach assumed not to be mapped to [-1., 1.] yet"""
        achieved_timed_goal = {
                "goal": achieved_goal,
                "delta_t_ach": [self.convert_time(delta_t_ach)]
                }
        return achieved_timed_goal

    @abstractmethod
    def map_to_goal(self, partial_obs):
        """Maps partial_obs to goal (assumed to be a deterministic map with image in goal_space)."""
        pass

    @abstractmethod
    def map_to_partial_obs(self, env_obs, parent_info):
            """Maps env_obs and parent_info to partial observation (assumed to be a deterministic map with image in partial_obs_space)."""
            pass

    @property
    def partial_obs_space(self):
        """Box space for partial observations (partial observation does not contain desired goal)."""
        return self._partial_obs_space

    @property
    def goal_space(self):
        """Box space containing goals of the subtask."""
        return self._goal_space

    @property
    def parent_action_space(self):
        """Space which contains actions the parent node is taking. 
        
        The action of a parent determines the the timed goal of the subtask via the method get_desired_timed_goal."""
        return self._parent_action_space

class BoxTGSubtaskSpec(TimedGoalSubtaskSpec):
    def __init__(self, goal_achievement_threshold, env, delta_t_max,  delta_t_min = 0., 
            delta_t_ach = None,  delta_t_comm = None, factorization = None):
        # Observation and goal spaces are all identical in this case
        self._space = get_obs_space_from_env(env)
        if factorization is None:
            self.factorization = [range(self._space.get_flat_dim())]
        else:
            self.factorization = factorization
        self._goal_achievement_threshold = listify(goal_achievement_threshold, 1)
        assert len(self.factorization) == len(self._goal_achievement_threshold)
        super(BoxTGSubtaskSpec, self).__init__(self._space, self._space, delta_t_ach, 
                delta_t_comm, delta_t_max = delta_t_max, delta_t_min = delta_t_min)

    def goal_achievement_criterion(self, achieved_goal, desired_goal, parent_info):
        # if the criterion in one of the subspaces is violated, the overall criterion is not met
        for factor, threshold in zip(self.factorization, self._goal_achievement_threshold):
            if np.linalg.norm(achieved_goal[factor] - desired_goal[factor]) >= threshold:
                return False
        return True

    def map_to_goal(self, partial_obs):
        # achieved goal is identical to partial observation
        return partial_obs

    def map_to_partial_obs(self, env_obs, parent_info):
        # all levels see full partial observation (but not env goal)
        return get_obs_from_gym(env_obs)

class BoxInfoHidingTGSubtaskSpec(TimedGoalSubtaskSpec):
    def __init__(self, partial_obs_indices, goal_indices, 
            goal_achievement_threshold, env, delta_t_max, delta_t_min = 0., 
            delta_t_ach = None, delta_t_comm = None, factorization = None):
        self._space = get_obs_space_from_env(env)
        if factorization is None:
            self.factorization = [range(self._space.get_flat_dim())]
        else:
            self.factorization = factorization
        self._goal_achievement_threshold = listify(goal_achievement_threshold, 1)
        assert len(self.factorization) == len(self._goal_achievement_threshold)
        self._partial_obs_indices = partial_obs_indices
        self._goal_indices = goal_indices

        self._space = self._space.get_flat_space(indices=partial_obs_indices)
        self._goal_space = self._space.get_flat_space(indices=goal_indices)
        super().__init__(self._goal_space, self._space, delta_t_ach, delta_t_comm, 
                delta_t_max=delta_t_max, delta_t_min=delta_t_min)

    def goal_achievement_criterion(self, achieved_goal, desired_goal, parent_info):
        # if the criterion in one of the subspaces is violated, the overall criterion is not met
        for factor, threshold in zip(self.factorization, self._goal_achievement_threshold):
            if np.linalg.norm(achieved_goal[factor] - desired_goal[factor]) >= threshold:
                return False
        return True

    def map_to_goal(self, partial_obs):
        # achieved goal is constructed from some of the dimensions of partial obs
        return partial_obs[self._goal_indices]

    def map_to_partial_obs(self, env_obs, parent_info):
        # partial observation is constructed by concatenating subset of 
        # components of partial env obs
        return get_obs_from_gym(env_obs)[self._partial_obs_indices]


"""class DictTGSubtaskSpec(TimedGoalSubtaskSpec):
    def __init__(self, goal_achievement_threshold, env, delta_t_ach = None,  delta_t_comm = None):
        assert isinstance(goal_achievement_threshold, dict)
        # Partial observation, parent action and goal spaces are all identical in this case
        self._space = DictSpace.get_from_gym_space(env.observation_space["observation"])
        self._goal_achievement_threshold = goal_achievement_threshold
        super(BoxTGSubtaskSpec, self).__init__(goal_space = self._space, delta_t_ach, delta_t_comm)

    def goal_achievement_criterion(self, achieved_goal, desired_goal, parent_info):
        # if the criterion in one of the subspaces is violated, the overall criterion is not met
        for key, threshold in self._goal_achievement_threshold.items():
            if np.linalg.norm(achieved_goal[key] - desired_goal[key]) >= threshold:
                return False
        return True"""

class DictInfoHidingTGSubtaskSpecBase(TimedGoalSubtaskSpec):
    """Subtask spec base with partial observation and goal dict space composed of 
    only some of the items of the env observation dict space.

    The partial observation is obtained by picking a subset of the keys of the env 
    partial observation. The achieved goal is obtained from the partial obs by 
    filtering the items with keys in goal_keys. See DefaultSPSubtaskSpec for 
    more details."""
    def __init__(self, partial_obs_keys, goal_keys, env, delta_t_ach = None, 
            delta_t_comm = None, delta_t_max = None, delta_t_min = 0.):
        self._partial_obs_keys = partial_obs_keys
        self._goal_keys = goal_keys
        orig_pos = get_obs_space_from_env(env)

        partial_obs_space = orig_pos.get_subspace(partial_obs_keys)
        goal_space = orig_pos.get_subspace(goal_keys)

        super(DictInfoHidingTGSubtaskSpecBase, self).__init__(goal_space, partial_obs_space, 
                delta_t_ach, delta_t_comm, delta_t_max, delta_t_min)

    def map_to_goal(self, partial_obs):
        return {key: value for key, value in partial_obs.items() if key in self._goal_keys}

    def map_to_partial_obs(self, env_obs, parent_info):
        return {key: value for key, value in get_obs_from_gym(env_obs).items() if key in self._partial_obs_keys}


class DictInfoHidingTGSubtaskSpec(DictInfoHidingTGSubtaskSpecBase):
    """Subtask spec with partial observation and goal dict space composed of 
    only some of the items of the env observation dict space.

    The desired goal is equal to the value corresponding to the goal key in 
    the parent action dict. See DictInfoHidingTGSubtaskSpecBase for more 
    details."""
    def __init__(self, goal_achievement_threshold, partial_obs_keys, 
            goal_keys, env, delta_t_ach=None,  delta_t_comm=None, 
            delta_t_max=None, delta_t_min=0., norms=None):
        assert isinstance(goal_achievement_threshold, dict)
        self._goal_achievement_threshold = goal_achievement_threshold
        if norms is None:
            self._norms = {key: None for key in goal_keys}
        else:
            self._norms = norms

        super(DictInfoHidingTGSubtaskSpec, self).__init__(partial_obs_keys, 
                goal_keys, env, delta_t_ach, delta_t_comm, delta_t_max, 
                delta_t_min)

    def goal_achievement_criterion(self, achieved_goal, desired_goal, parent_info):
        # if the criterion in one of the subspaces is violated, the overall criterion is not met
        for key, threshold in self._goal_achievement_threshold.items():
            if (np.linalg.norm(achieved_goal[key] - desired_goal[key],ord=self._norms[key])
                    >= threshold):
                return False
        return True

class DictInfoHidingTolTGSubtaskSpec(DictInfoHidingTGSubtaskSpecBase):
    """Subtask spec with partial observation and goal dict space composed of 
    only some of the items of the env observation dict space. Parent action 
    determines not only goal itself but also tolerance for achievement.

    The desired goal is equal to the value corresponding to the goal key in 
    the parent action dict. See DictInfoHidingTGSubtaskSpecBase for more 
    details."""
    def __init__(self, partial_obs_keys, goal_keys, env, max_goal_thresholds = None, 
            delta_t_ach = None, delta_t_comm = None, delta_t_max = None, delta_t_min = 0.):
        super(DictInfoHidingTolTGSubtaskSpec, self).__init__(partial_obs_keys, 
                goal_keys, env, delta_t_ach, delta_t_comm, delta_t_max, delta_t_min)


        # add goal tolerance to parent action space so as higher level can 
        # choose goal achievement tolerance
        space_dict = self._parent_action_space._space_dict
        if max_goal_thresholds is None:
            max_goal_thresholds = {key: 1. for key in goal_keys}
        else:
            assert max_goal_thresholds.keys() == goal_keys
        self._goal_tol_space = DictSpace({key: BoxSpace(low = [0.], high = [value]) 
                for key, value in max_goal_thresholds.items()})
        space_dict["goal_tol"] = self._goal_tol_space 
        self._parent_action_space = DictSpace(space_dict)
    
    def goal_achievement_criterion(self, achieved_goal, desired_goal, parent_info):
        # if the criterion in one of the subspaces is violated, the overall criterion is not met
        for key, threshold in parent_info.action["goal_tol"].items():
            if np.linalg.norm(achieved_goal[key] - desired_goal[key]) >= threshold:
                return False
        return True

    def get_achieved_timed_goal_dict(self, achieved_goal, delta_t_ach, parent_info):
        """delta_t_ach assumed not to be mapped to [-1., 1.] yet"""
        atg = super(DictInfoHidingTolTGSubtaskSpec, self).get_achieved_timed_goal_dict(achieved_goal, delta_t_ach, 
                parent_info)
        atg["goal_tol"] = parent_info.action["goal_tol"]
        return atg



from abc import ABC, abstractmethod

class Space(ABC):

    def __init__(self):
        super().__init__()

    @abstractmethod
    def get_flat_space(self):
        pass

    @abstractmethod
    def flatten_value(self, value):
        pass

    @abstractmethod
    def get_flat_dim(self):
        """Get dimension of flat space associated to this space."""
        pass

    @abstractmethod
    def unflatten_value(self, flat_value):
        """Maps a vector from the space returned by flatten_value back to this space."""
        pass

    @abstractmethod
    def contains(self, value):
        pass

    @abstractmethod
    def sample(self):
        pass

    @abstractmethod
    def __eq__(self, other):
        pass

    @staticmethod
    @abstractmethod
    def concatenate(*args):
        pass

    @staticmethod
    @abstractmethod
    def get_from_gym_space(gym_space):
        pass

    @abstractmethod
    def __str__(self):
        pass



from abc import ABC, abstractmethod

class SubtaskSpec(ABC):
    """Base class for subtask specifications."""

    def __init__(self):
        super().__init__()
        self._aux_rewards = []

    def add_aux_reward(self, aux_reward):
        """Add auxiliary reward."""
        self._aux_rewards.append(aux_reward)

    def get_aux_rewards(self):
        """Get list of auxiliary rewards."""
        return self._aux_rewards



from gym import GoalEnv
from gym import spaces

from ..spaces import space_from_gym_space

def get_obs_space_from_env(env):
    if (isinstance(env.observation_space, spaces.Dict) 
        and "observation" in env.observation_space.spaces):
        return space_from_gym_space(env.observation_space["observation"])
    else:
        return space_from_gym_space(env.observation_space)

def get_obs_from_gym(gym_obs):
    if isinstance(gym_obs, dict) and "observation" in gym_obs:
        return gym_obs["observation"]
    else:
        return gym_obs



from gym import spaces as gym_spaces

from . import BoxSpace, TupleSpace, MultiBinarySpace, DictSpace

gym_conv_dict = {
        gym_spaces.Box: BoxSpace,
        gym_spaces.Dict: DictSpace,
        gym_spaces.MultiBinary: MultiBinarySpace
        }



import os 
import time
from datetime import datetime
import json

from gym import GoalEnv
import numpy as np
from tianshou.data.utils.converter import to_numpy
import torch
from torch.utils.tensorboard import SummaryWriter

from ..data import EnvInfo, SessInfo
from ..logging import CSVLogger, read_csv_log


class Session:

    def __init__(self, graph, env):
        self.graph = graph
        self.env = env

        # detect whether environment is goal-based
        if isinstance(env, GoalEnv):
            self._goal_based_env = True
        else:
            self._goal_based_env = False

    def _process_env_observation(self, obs):
        # convert lists and tuples into numpy arrays
        return to_numpy(obs)

    def step_env_and_graph(self, env_info, sess_info, render, render_frequency, episode, total_step_multiplier = 1.):
            # get atomic action from forward pass through graph
            action = self.graph.get_atomic_action(env_info.new_obs, sess_info, testing = sess_info.testing)

            # step environment
            new_raw_obs, env_rew, env_done, info = self.env.step(action)
            new_env_obs = self._process_env_observation(new_raw_obs)
            new_env_info = EnvInfo(
                    new_obs = new_env_obs, 
                    reward = env_rew,
                    done = env_done,
                    info = info)

            sess_info.total_step += 1*total_step_multiplier
            sess_info.ep_step += 1

            # render 
            if render and episode % render_frequency == 0:
                self.env.render()

            # register experience in two backward passes through graph
            graph_done = self.graph.register_experience(new_env_info, sess_info)

            return new_env_info, graph_done

    def run(self, n_steps = None, n_episodes = None, max_runtime = None, learn = True, render = False, 
            render_frequency = 10, test = False, test_frequency = 10, test_episodes = 1, 
            test_render = False, test_render_frequency = 1, verbose = 1, 
            tensorboard_logdir = None, csv_logdir = None, run_name = None, 
            append_run_name_to_log_paths = True, torch_num_threads = None, 
            cb_after_train_episode = None, total_step_init = 0, append_to_logfiles = False, 
            success_reward = None, train=True):
        """Run reinforcement learning session with policy induced by graph in env.
        Args:
            n_steps (int or None): Number of training timesteps to take. Can be None if 
                n_episodes is not None. In this case the argument is ignored. 
                One of the arguments n_steps and n_episodes has to be an int 
                and one None.
            n_episodes (int or None): Number of training episodes to run. Can be None if 
                n_steps is not None. In this case the argument is ignored. 
                One of the arguments n_steps and n_episodes has to be an int 
                and one None.
            success_reward (float or None): If this reward is encountered 
                during an episode, it is counted as a successful episode."""

        assert {type(n_steps), type(n_episodes)} == {int, type(None)}, \
                "Please provide either a number of time steps to take or \
                a number of episodes to run (in n_steps or n_episodes)."


        if torch_num_threads is not None:
            torch.set_num_threads(torch_num_threads)
            print(f"Using {torch_num_threads} threads in pytorch.")

        if verbose >= 1:
            print("-------------------------")
            print("Running session with")
            print("Env class: {}".format(type(self.env).__name__))
            print("Graph class: {}".format(type(self.graph).__name__))
            if n_steps is not None:
                print("Number of timesteps: {}".format(n_steps))
            if n_episodes is not None:
                print("Number of episodes: {}".format(n_episodes))
            print("Learn: {}".format(learn))
            print("-------------------------\n")

        if run_name is None:
            current_time = datetime.now()
            run_name = current_time.strftime("%Y_%m_%d_%H_%M")

        if tensorboard_logdir is not None:
            if append_run_name_to_log_paths:
                tb_logdir = os.path.join(tensorboard_logdir, run_name, "tensorboard")
            else:
                tb_logdir = tensorboard_logdir
            writer = SummaryWriter(tb_logdir)
            self.graph.set_tensorboard_writer(writer)
        else:
            writer = None

        if csv_logdir is not None:
            if append_run_name_to_log_paths:
                csv_logdir_run = os.path.join(csv_logdir, run_name, "log")
            else:
                csv_logdir_run = csv_logdir
            os.makedirs(csv_logdir_run, exist_ok = True)
            self.graph.create_logfiles(csv_logdir_run, append = append_to_logfiles)
            fieldnames_train = ["return", "length", "time"]
            fieldnames_test = ["return", "length", "step", "time"]
            if success_reward is not None:
                fieldnames_train.append("success")
                fieldnames_test.append("success")
                fieldnames_test.append("return_success")

            logger_train = CSVLogger(os.path.join(csv_logdir_run, "session_train.csv"), 
                    fieldnames_train, append = append_to_logfiles)
            logger_test = CSVLogger(os.path.join(csv_logdir_run, "session_test.csv"), 
                    fieldnames_test, append = append_to_logfiles)
        else:
            logger_train = None
            logger_test = None

        ep_return = 0.
        test_episode = 0
        episode = 0
        success = False

        env_info = EnvInfo(
                new_obs = self._process_env_observation(self.env.reset()), 
                done = True)
        sess_info = SessInfo(
                ep_step = 0,
                total_step = total_step_init,
                learn = learn,
                testing = False)
        self.graph.reset()

        t_start = time.process_time()

        while (n_steps is None or sess_info.total_step < n_steps) and \
                (n_episodes is None or episode < n_episodes) and \
                (max_runtime is None or time.process_time() - t_start < max_runtime):

            if verbose >= 1 and (env_info.done or graph_done):
                print("\nEpisode {}:".format(episode))

            if verbose >= 2:
                print("Step {}:".format(sess_info.total_step))

            if train:
                env_info, graph_done = self.step_env_and_graph(env_info, sess_info, 
                        render, render_frequency, episode)
                if success_reward is not None and env_info.reward == success_reward:
                    success = True
                    print("Successful episode!")
                ep_return += env_info.reward
            else:
                env_info = EnvInfo()
                graph_done = False

            # end of a training episode
            if not train or env_info.done or graph_done:
                if verbose >= 1:
                    print("Return: {}".format(ep_return))
                    print("env_done: {}".format(env_info.done))
                    print("graph_done: {}".format(graph_done))
                    print("")

                if writer is not None:
                    writer.add_scalar("env/return/train", ep_return, sess_info.total_step)
                    writer.add_scalar("env/ep_length/train", sess_info.ep_step, sess_info.total_step)
                if logger_train is not None:
                    row_dict = {
                            "return": ep_return, 
                            "length": sess_info.ep_step, 
                            "time": logger_train.time_passed()
                            }
                    if success_reward is not None:
                        row_dict["success"] = int(success)
                    logger_train.log(row_dict)

                if cb_after_train_episode is not None:
                    cb_after_train_episode(self.graph, sess_info, ep_return, graph_done)

                # learn
                if sess_info.learn:
                    self.graph.learn(sess_info)

                episode += 1
                success = False
                ep_return = 0

                # testing
                if test and episode % test_frequency == 0:
                    if verbose >= 1:
                        print("Testing")
                    cumulative_env_return_test = 0
                    # return of successful episodes
                    cumulative_env_return_test_success = 0
                    cumulative_env_length_test = 0
                    cumulative_success_test = 0
                    sess_info_test = SessInfo(
                            ep_step = 0,
                            total_step = sess_info.total_step,
                            learn = False,
                            testing = True)
                    for test_episode in range(test_episodes):
                        env_info = EnvInfo(
                                new_obs = self._process_env_observation(self.env.reset()), 
                                done = False)
                        self.graph.reset()
                        env_return_test = 0
                        success_test = False
                        sess_info_test.ep_step = 0
                        while True:
                            env_info, graph_done = self.step_env_and_graph(env_info, sess_info_test, 
                                    test_render, test_render_frequency, test_episode, 1./test_episodes)

                            if (success_reward is not None and 
                                    env_info.reward == success_reward):
                                success_test = True
                            env_return_test += env_info.reward

                            if env_info.done or graph_done:
                                if verbose >= 1:
                                    print("Return: {}".format(env_return_test))
                                    print("env_done: {}".format(env_info.done))
                                    print("graph_done: {}\n".format(graph_done))
                                cumulative_env_return_test += env_return_test
                                cumulative_env_length_test += sess_info_test.ep_step
                                if success_test:
                                    cumulative_env_return_test_success += env_return_test
                                cumulative_success_test += float(success_test)
                                break

                    avg_return_test = cumulative_env_return_test/test_episodes
                    avg_length_test = cumulative_env_length_test/test_episodes
                    avg_success_test = cumulative_success_test/test_episodes
                    if cumulative_success_test > 0.:
                        avg_return_test_success = (cumulative_env_return_test_success/
                                cumulative_success_test)
                    else:
                        avg_return_test_success = np.nan
                    if writer is not None:
                        # fractional steps for testing to keep logging consistent
                        writer.add_scalar("env/return/test", avg_return_test, 
                                sess_info.total_step)
                        writer.add_scalar("env/ep_length/test", avg_length_test, 
                                sess_info.total_step)
                    if logger_test is not None:
                        row_dict = {
                                "return": avg_return_test, 
                                "length": avg_length_test, 
                                "step": sess_info.total_step, 
                                "time": logger_test.time_passed()
                                }
                        if success_reward is not None:
                            row_dict["success"] = avg_success_test
                            row_dict["return_success"] = avg_return_test_success
                        logger_test.log(row_dict)

                # reset env_info and graph for next training episode
                env_info = EnvInfo(
                        new_obs = self._process_env_observation(self.env.reset()), 
                        done = True)
                sess_info.ep_step = 0
                self.graph.reset()

        t_elapsed = time.process_time() - t_start

        session_props = {}
        for arg_name in self.run.__code__.co_varnames[:self.run.__code__.co_argcount]:
            if arg_name != "self" and type(locals()[arg_name]) in [str, bool, int, float]:
                session_props[arg_name] = locals()[arg_name]
        session_props["t_elapsed"] = t_elapsed
        session_props["timed_out"] = max_runtime is not None and t_elapsed >= max_runtime
        session_props["total_step"] = sess_info.total_step
        # log session properties (all arguments to run, elapsed time)
        if csv_logdir is not None:
            session_file = open(os.path.join(csv_logdir_run, "session.json"), "w")
            json.dump(session_props, session_file, indent = 4)

        return session_props

    @classmethod
    def read_logfiles(cls, log_path):
        data = {
                "train": read_csv_log(os.path.join(log_path, "session_train.csv")), 
                "test": read_csv_log(os.path.join(log_path, "session_test.csv"))
                }
        # reconstruct step at end of training episode from episode length
        current_step = 0
        steps = []
        for l in data["train"]["length"]:
            current_step += l
            steps.append(current_step)
        data["train"]["step"] = steps
        return data



from copy import deepcopy
from math import floor

import numpy as np
import torch

from ..data import FlatTransition
from . import OffPolicyAlgorithm
from ..subtasks.timed_goal_subtask_specs import TimedGoal

class HiTS(OffPolicyAlgorithm):
    """Hierarchical reinforcement learning with Timed Subgoals."""

    # sampling strategy "episode" does not make sense for timed goals
    supported_goal_sampling_strategies = {"future", "final"}

    def __init__(self, name, flat_algo_name, flat_algo_kwargs, model, 
            child_failure_penalty, check_status, convert_time, goal_sampling_strategy = "future", 
            buffer_size = 100000, batch_size = 128, n_hindsight_goals = 4, testing_fraction = 0.3, 
            fully_random_fraction = 0.2, bootstrap_testing_transitions = True, 
            use_normal_trans_for_testing = False, use_testing_transitions = True, 
            learn_from_deterministic_episodes = True, log_q_values = True, learning_starts = 0, 
            grad_steps_per_env_step=1, delta_t_min=0.):
            """
            Args:
                check_status (callable): Maps achieved_goal, 
                    desired_timed_goal, step and parent_info to 
                    * achieved: indicates whether timed goal has been achieved
                    * reward: reward upon arriving in achieved_goal
                    * ach_time_up: whether desired time until achievement ran out
                    * comm_time_up: whether commitment time ran out
                delta_t_min: Minimum delta_t. Important for sampling hindsight
                    timed subgoals with HER.
                """
            super(HiTS, self).__init__(name, flat_algo_name, model, fully_random_fraction, 
                    flat_algo_kwargs, buffer_size, batch_size, learning_starts,
                    grad_steps_per_env_step)

            assert goal_sampling_strategy in self.supported_goal_sampling_strategies, \
                    "Goal sampling strategy {} not supported.".format(goal_sampling_strategy)


            self._child_failure_penalty = child_failure_penalty
            self._goal_sampling_strategy = goal_sampling_strategy
            self._check_status = check_status
            self._convert_time = convert_time
            self._n_hindsight_goals = n_hindsight_goals
            self._testing_fraction = testing_fraction
            self._bootstrap_testing_transitions = bootstrap_testing_transitions

            self._use_normal_trans_for_testing = use_normal_trans_for_testing
            self._use_testing_transitions = use_testing_transitions
            self._learn_from_deterministic_episodes = learn_from_deterministic_episodes
            self._log_q_values = log_q_values
            self._delta_t_min = delta_t_min

            self._verbose = False

    def _sample_achieved_timed_goals(self, current_index, episode_start):
        n_transitions = len(self._episode_transitions)
        if self._goal_sampling_strategy == "future":
            low = current_index
            # Adapt number of goals to current_index in a way which is 
            # equivalent to sampling indices uniformly from whole episode
            # and rejecting those which lie in the past. Factor 2 to have
            # number of actual hindsight goals approx. equal to _n_hindsight_goals
            # in expectation.
            n_goals = 2.*(n_transitions - low)/n_transitions*self._n_hindsight_goals
            frac = n_goals - floor(n_goals)
            n_goals_int = floor(n_goals) + (1 if np.random.rand() <= frac else 0)
            indices = np.random.randint(low=low, high=n_transitions, size=n_goals_int)
        elif self._goal_sampling_strategy == "final":
            indices = [-1]

        goals = []
        # time of current new_obs, i.e., time after transition (in env steps)
        current_t = self._episode_transitions[current_index].subtask_tr.info["t"]
        for i in indices:
            # sample delta_t_ach uniformly from interval which would have run out in the 
            # step corresponding to hindsight goal
            rand_shift = np.random.rand()
            goals.append({
                "desired_goal": self._episode_transitions[i].subtask_tr.info["achieved_generalized_goal"]["goal"], 
                "delta_t_ach_obs": [self._convert_time(self._episode_transitions[i].subtask_tr.info["t"] - (current_t - 1) - rand_shift)], 
                "delta_t_ach_new_obs": [self._convert_time(self._episode_transitions[i].subtask_tr.info["t"] - current_t - rand_shift)]
                })
        return goals

    def _add_experience_to_flat_algo(self, parent_info, deterministic_episode, node_is_sink, sess_info):
        if self._learn_from_deterministic_episodes or not deterministic_episode:

            # add transitions of this episode to replay buffer of flat RL algorithm
            # (by applying hindsight goal and action manipulations)
            for trans_index, tr in enumerate(self._episode_transitions):

                # unaltered flat transition
                f_trans_0 = FlatTransition(
                        obs = tr.subtask_tr.obs, 
                        action = tr.subtask_tr.action,
                        reward = tr.subtask_tr.reward,
                        new_obs = tr.subtask_tr.new_obs,
                        done = tr.subtask_tr.info["ach_time_up"])

                # if the node is a sink, do not attempt to manipulate action
                # in hindsight and add original transition to replay buffer
                if node_is_sink:
                    f_trans_base = f_trans_0
                    self._add_to_flat_replay_buffer(f_trans_0)

                # if the node is not a sink consider testing transitions
                # and hindsight action transitions
                else:
                    # did the child node use a deterministic version of its policy?
                    # (testing transition)
                    testing_transition = tr.algo_info["child_be_deterministic"]

                    # boolean indicating whether child achieved subgoal
                    did_child_achieve_subgoal = tr.child_feedback["has_achieved"]

                    # add testing transitions (if enabled)
                    # Optionally, also transitions generated by stochastic 
                    # lower level are used as testing transitions.
                    # Only add transition with penalty if child node failed 
                    # to achieve its subgoal, otherwise do not add any transition.
                    if self._use_testing_transitions \
                       and (testing_transition or self._use_normal_trans_for_testing) \
                       and not did_child_achieve_subgoal:
                        f_trans_testing = deepcopy(f_trans_0)
                        f_trans_testing.reward = self._child_failure_penalty
                        if not self._bootstrap_testing_transitions:
                            # Have to set done to True in these transitions 
                            # (according to HAC paper and accompanying code).
                            f_trans_testing.done = True
                        if self._verbose:
                            print("testing transition in {}\n".format(self.name) 
                                    + str(f_trans_testing))
                        self._add_to_flat_replay_buffer(f_trans_testing)

                    # hindsight action transition
                    # If child node achieved desired goal use original action.
                    # If not, use subgoal the child achieved as action.
                    f_trans_hindsight_action = deepcopy(f_trans_0)
                    if not did_child_achieve_subgoal:
                        f_trans_hindsight_action.action = deepcopy(f_trans_hindsight_action.action)
                        f_trans_hindsight_action.action["goal"] = tr.child_feedback["achieved_generalized_goal"]["goal"]
                        # TODO: In principle, reward could depend on action, so have to recompute
                        if self._verbose:
                            print("hindsight action transition in {}\n".format(self.name) 
                                    + str(f_trans_hindsight_action))
                    self._add_to_flat_replay_buffer(f_trans_hindsight_action)
                    f_trans_base = f_trans_hindsight_action

                # hindsight goal transitions (based on hindsight action 
                # transition or original transition in case of a sink node)
                achieved_timed_goals = self._sample_achieved_timed_goals(trans_index, parent_info.step)
                for hindsight_tg in achieved_timed_goals:
                    f_trans_hindsight_goal = deepcopy(f_trans_base)
                    f_trans_hindsight_goal.obs = {
                            "partial_observation": tr.subtask_tr.obs["partial_observation"], 
                            "desired_goal": hindsight_tg["desired_goal"], 
                            "delta_t_ach": hindsight_tg["delta_t_ach_obs"]}
                    f_trans_hindsight_goal.new_obs = {
                            "partial_observation": tr.subtask_tr.new_obs["partial_observation"], 
                            "desired_goal": hindsight_tg["desired_goal"], 
                            "delta_t_ach": hindsight_tg["delta_t_ach_new_obs"]}
                    h_tg = TimedGoal(
                            goal = hindsight_tg["desired_goal"], 
                            delta_t_ach = hindsight_tg["delta_t_ach_new_obs"][0], 
                            delta_t_comm = hindsight_tg["delta_t_ach_new_obs"][0])
                    achieved, reward, ach_time_up, _ = self._check_status(
                            achieved_goal = tr.subtask_tr.info["achieved_generalized_goal"]["goal"], 
                            desired_tg = h_tg, 
                            obs = f_trans_hindsight_goal.obs, 
                            action = f_trans_hindsight_goal.action, 
                            parent_info = parent_info,
                            env_info = tr.env_info) 
                    f_trans_hindsight_goal.done = ach_time_up
                    f_trans_hindsight_goal.reward = reward
                    self._add_to_flat_replay_buffer(f_trans_hindsight_goal)

        self._episode_transitions.clear()

    def get_algo_info(self, env_obs, parent_info):
        # child_be_deterministic instructs the children to be deterministic whereas
        # is_deterministic implies that this node is supposed to be deterministic
        if parent_info is not None:
            is_deterministic = parent_info.algo_info["child_be_deterministic"]
            child_be_deterministic = is_deterministic
        else:
            is_deterministic = False
            child_be_deterministic = False

        # if child_be_deterministic is true, all children have to use deterministic policies
        # until this node or an active parent gets back control (testing transition)
        child_be_deterministic = child_be_deterministic or np.random.rand() < self._testing_fraction
        new_algo_info = {
                "is_deterministic": is_deterministic, 
                "child_be_deterministic": child_be_deterministic
                }
        return new_algo_info



from copy import copy
import os

import numpy as np

from . import Subtask
from ..spaces import DictSpace, BoxSpace, space_from_gym_space
from ..data import SubtaskTransition
from ..utils import listify
from ..logging import CSVLogger

class TimedGoalSubtask(Subtask):
    """Subtask is to achieve a goal at a given point in time."""

    def __init__(self, name, subtask_spec):
            """Args:
                subtask_spec (TimedGoalSubtaskSpec): Specification of the timed goal subtask.
            """
            super(TimedGoalSubtask, self).__init__(name, subtask_spec)

            self.task_spec = subtask_spec

            self._observation_space = DictSpace({
                "partial_observation": subtask_spec.partial_obs_space, 
                "desired_goal": subtask_spec.goal_space,
                "delta_t_ach": BoxSpace(low = [-1.], high = [1.]),
                })

            self._n_actions_taken = 0

            self.logger_test = None
            self.logger_train = None
            self.logfiles = {}

    @property
    def observation_space(self):
        return self._observation_space

    @property
    def parent_action_space(self):
        return self.task_spec.parent_action_space

    def reset(self):
        self._n_actions_taken = 0

    def get_observation(self, env_obs, parent_info, sess_info):
        partial_obs = self.task_spec.map_to_partial_obs(env_obs, parent_info)
        desired_tg = self.task_spec.get_desired_timed_goal(env_obs, parent_info)

        full_obs = {
                "partial_observation": partial_obs, 
                "desired_goal": desired_tg.goal, 
                # policy sees desired time until achievement relative to current time step
                "delta_t_ach": [self.task_spec.convert_time(self.task_spec.unconvert_time(desired_tg.delta_t_ach) - (sess_info.ep_step - parent_info.step))]
                }
        return full_obs


    def check_status(self, achieved_goal, desired_tg, obs, action, parent_info, env_info):
        """Assumes that the desired_tg has delta_t attributes in [-1, 1] space and not in units of env steps."""
        ach_time_up = self.task_spec.unconvert_time(desired_tg.delta_t_ach) <= 0.
        # timed goal can only be achieved when the time is up
        if  ach_time_up and \
        self.task_spec.goal_achievement_criterion(achieved_goal, desired_tg.goal, parent_info):
            reward = 1.
            achieved = True
        else:
            reward = 0.             
            achieved = False
        # running out of commitment time 
        # NOTE: This is always false if commitment time is NaN
        comm_time_up = self.task_spec.unconvert_time(desired_tg.delta_t_comm) <= 0.

        # add auxiliary rewards
        if obs is not None and action is not None:
            reward += self.get_aux_rewards(obs, action, env_info.reward)
            
        return achieved, reward, ach_time_up, comm_time_up

    def _check_status_convenience(self, achieved_goal, obs, action, parent_info, sess_info, env_info):
        """Convenience version of check status also does the conversion of delta_t into [-1, 1] interval. 
        
        It assumes that the delta_t in parent_info has not been updated since the emission of the parent_info by 
        the parent. Hence, the elapsed time is subtracted from this outdated delta_t."""
        desired_tg = self.get_updated_desired_tg_in_steps(env_info.new_obs, parent_info, sess_info)
        desired_tg.delta_t_ach = self.task_spec.convert_time(desired_tg.delta_t_ach)
        desired_tg.delta_t_comm = self.task_spec.convert_time(desired_tg.delta_t_comm)
        
        return self.check_status(achieved_goal, desired_tg, obs, action, parent_info, env_info)


    def check_interruption(self, env_info, new_subtask_obs, parent_info, sess_info):
        super().check_interruption(env_info, new_subtask_obs, parent_info, sess_info)
        new_env_obs = env_info.new_obs
        new_partial_obs = self.task_spec.map_to_partial_obs(new_env_obs, parent_info)
        achieved_goal = self.task_spec.map_to_goal(new_partial_obs)
        _, _, ach_time_up, comm_time_up = self._check_status_convenience(
                achieved_goal, None, None, parent_info, sess_info, env_info)
        return ach_time_up or comm_time_up

    def evaluate_transition(self, env_obs, env_info, subtask_trans, parent_info, algo_info, sess_info):
        new_partial_obs = self.task_spec.map_to_partial_obs(env_info.new_obs, parent_info)
        achieved_goal = self.task_spec.map_to_goal(new_partial_obs)
        achieved, reward, ach_time_up, comm_time_up = self._check_status_convenience(achieved_goal, 
                subtask_trans.obs, subtask_trans.action, parent_info, sess_info, env_info)

        # subtask ended if subgoal achieved or running out of commitment time
        ended = ach_time_up or comm_time_up

        self._n_actions_taken += 1
        if ended:
            n_actions_taken = self._n_actions_taken
            self._n_actions_taken = 0

        # sample delta_t_ach for achieved timed goal from uniform distribution over times that would have run out
        # in this env step
        achieved_timed_goal = self.task_spec.get_achieved_timed_goal_dict(
                achieved_goal = achieved_goal, 
                delta_t_ach = sess_info.ep_step - parent_info.step - np.random.rand(), 
                parent_info = parent_info)

        # NOTE: The boolean subtask_ended only indicates whether the subtask ended, not that the goal was reached!
        # Whether the goal was reached is encoded in the key "has_achieved" in info.
        info = {
                "has_achieved": achieved,
                "comm_time_up": comm_time_up, 
                "ach_time_up": ach_time_up, 
                "achieved_generalized_goal": achieved_timed_goal
                }
        feedback = copy(info)
        #info["delta_t_comm"] = desired_tg.delta_t_comm - (sess_info.ep_step - parent_info.step)
        info["t"] = sess_info.ep_step

        complete_subtask_trans = copy(subtask_trans)
        complete_subtask_trans.reward = reward
        complete_subtask_trans.ended = ended
        complete_subtask_trans.info = info

        # tensorboard logging
        if self._tb_writer is not None and ach_time_up:
                mode = "test" if sess_info.testing else "train"
                self._tb_writer.add_scalar(f"{self.name}/{mode}/subgoal_achieved", int(achieved), sess_info.total_step)
                self._tb_writer.add_scalar(f"{self.name}/{mode}/n_actions", n_actions_taken, sess_info.total_step)
        # csv logging
        if ach_time_up and self.logger_train is not None and self.logger_test is not None:
            row_dict = {
                    "achieved": int(achieved), 
                    "n_actions": n_actions_taken, 
                    "step": sess_info.total_step, 
                    "time": self.logger_test.time_passed() if sess_info.testing else self.logger_train.time_passed()
                    }
            if sess_info.testing:
                self.logger_test.log(row_dict)
            elif self.logger_train:
                self.logger_train.log(row_dict)

        return complete_subtask_trans, feedback
    
    def get_updated_desired_tg_in_steps(self, env_obs, parent_info, sess_info):
        """Returns desired timed goal in global time (i.e. env steps)."""
        desired_tg = self.task_spec.get_desired_timed_goal(env_obs, parent_info)

        elapsed_time = sess_info.ep_step - parent_info.step

        desired_tg.delta_t_ach = self.task_spec.unconvert_time(desired_tg.delta_t_ach) - elapsed_time
        desired_tg.delta_t_comm = self.task_spec.unconvert_time(desired_tg.delta_t_comm) - elapsed_time

        return desired_tg

    def create_logfiles(self, logdir, append):
        if logdir is not None:
            logfile_test = os.path.join(logdir, self.name  + "_test.csv")
            logfile_train = os.path.join(logdir, self.name  + "_train.csv")
            self.logger_test = CSVLogger(logfile_test, ("achieved", "n_actions", "step", "time"), append)
            self.logger_train = CSVLogger(logfile_train, ("achieved", "n_actions", "step", "time"), append)
            self.logfiles["test"] = logfile_test
            self.logfiles["train"] = logfile_train

    def get_logfiles(self):
        return self.logfiles





from . import Node
from ..subtasks import (
    ShortestPathSubtask, ShortestPathSubtaskSpec, ReturnMaximSubtask, ReturnMaximSubtaskSpec
)
from ..algorithms import HAC

class HACNode(Node):

    def __init__(self, name, parents, subtask_spec, HAC_kwargs, tb_writer=None, interruption_policy=None):

        # create subtask (shortest path and reward maximization subtasks are supported)
        if isinstance(subtask_spec, ShortestPathSubtaskSpec):
            subtask = ShortestPathSubtask(name + "_subtask", subtask_spec)
        elif isinstance(subtask_spec, ReturnMaximSubtaskSpec):
            subtask = ReturnMaximSubtask(name + "_subtask", subtask_spec)
        else:
            raise NotImplementedError("An HAC node supports only a shortest path and a return maximization subtask.")

        # create algorithm
        check_achievement = subtask.check_achievement if hasattr(subtask, "check_achievement") else None 
        algorithm = HAC(
                name = name + "_algorithm",
                check_achievement = check_achievement,
                use_normal_trans_for_testing = False,
                use_testing_transitions = True,  
                learn_from_deterministic_episodes = True, 
                **HAC_kwargs)

        # policy creation is done via algorithm because it does the sampling part
        # of generating an action
        policy_class = algorithm.get_policy

        super(HACNode, self).__init__(name, policy_class, subtask, algorithm, parents, interruption_policy)

    def get_action(self, env_obs, parent_info, parent, sess_info, testing):
        if hasattr(self.subtask.task_spec, "get_desired_goal"):
            self.current_goal = self.subtask.task_spec.get_desired_goal(env_obs, parent_info)
        return super(HACNode, self).get_action(env_obs, parent_info, parent, sess_info, testing)



from abc import ABC, abstractmethod

from ..logging import read_csv_log

class Subtask(ABC):

    def __init__(self, name, subtask_spec):
        super().__init__()
        self.name = name
        self._tb_writer = None
        self._aux_rewards = subtask_spec.get_aux_rewards()

    @property
    @abstractmethod
    def observation_space(self):
        pass

    @property
    @abstractmethod
    def parent_action_space(self):
        pass

    @abstractmethod
    def reset(self):
        """Executed when node is reset."""
        pass

    @abstractmethod
    def get_observation(self, env_obs, parent_info):
        pass

    @abstractmethod
    def check_interruption(self, env_info, new_subtask_obs, parent_info, sess_info):
        for aux_rew in self._aux_rewards:
            aux_rew.update(env_info)

    @abstractmethod
    def evaluate_transition(self, env_obs, env_info, subtask_trans, parent_info, algo_info, sess_info):
        pass

    def get_aux_rewards(self, obs, action, env_reward):
        """Get sum of auxiliary rewards."""

        r = 0
        for aux_rew in self._aux_rewards:
            r += aux_rew(obs, action, env_reward)
        return r

    def add_aux_reward(self, aux_reward):
        self._aux_rewards.append(aux_reward)

    def set_tensorboard_writer(self, tb_writer):
        self._tb_writer = tb_writer

    def create_logfiles(self, logdir, append = False):
        pass

    @classmethod
    def read_logfiles(cls, logfiles):
        data = {}
        for name, path in logfiles.items():
            data[name] = read_csv_log(path)
        return data

    def get_logfiles(self):
        return None




from abc import ABC, abstractmethod

import torch

from .model import Model
from ..networks import MLPNetwork, ConcatNetwork, SquashNetwork, NegativeNetwork

class MLPActorCriticModel(Model):
    def __init__(self, hidden_layers_actor, hidden_layers_critics, activation_fns_actor, 
            activation_fns_critics, n_critics, learning_rate_actor, learning_rate_critics, 
            device, squash_critics = False, low = None, high = None, force_negative = False):
        self._hidden_layers_actor = hidden_layers_actor
        self._hidden_layers_critics = hidden_layers_critics
        self._activation_fns_actor = activation_fns_actor
        self._activation_fns_critics = activation_fns_critics
        self._n_critics = n_critics
        self._learning_rate_actor = learning_rate_actor
        self._learning_rate_critics = learning_rate_critics
        self._device = device
        self._squash_critics = squash_critics
        self._low = low
        self._high = high
        self._force_negative = force_negative

    @abstractmethod
    def create(self, state_dim, action_dim):
        self._actor_stump = MLPNetwork(state_dim, self._hidden_layers_actor, self._activation_fns_actor, 
                device = self._device)

        self._critics = []
        self._critic_optims = []
        for _ in range(self._n_critics):
            critic = MLPNetwork(state_dim + action_dim, self._hidden_layers_critics + [1], 
                    self._activation_fns_critics + [None])
            if self._force_negative:
                critic = NegativeNetwork(critic, self._device)
            elif self._squash_critics:
                critic = SquashNetwork(critic, self._low, self._high, self._device)
            critic = ConcatNetwork(critic, device = self._device)
            self._critics.append(critic)
            self._critic_optims.append(torch.optim.Adam(critic.parameters(), lr = self._learning_rate_critics))

    def set_up_clipping(self, actor_threshold, critic_threshold):
        """Set up hooks for gradient clipping."""

        def rescale(grad, threshold):
            norm = torch.norm(grad)
            if norm > threshold:
                return threshold*grad/norm
            else:
                return grad

        if critic_threshold is not None:
            for network in self._critics:
                for p in network.parameters():
                    p.register_hook(lambda grad: rescale(grad, critic_threshold))

        if actor_threshold is not None:
            for p in self._actor.parameters():
                    p.register_hook(lambda grad: rescale(grad, actor_threshold))


    def set_learning_rate_critics(self, lr):
        self._learning_rate_critics = lr
        for optim in self._critic_optims:
            for g in optim.param_groups:
                g["lr"] = lr

    def set_learning_rate_actor(self, lr):
        self._learning_rate_actor = lr
        for g in self._actor_optim.param_groups:
            g["lr"] = lr

    @property
    def actor(self):
        return self._actor


    @property
    def critics(self):
        return self._critics


    @property
    def critic_optims(self):
        return self._critic_optims


    def get_parameters(self):
        """Returns dictionary containing all learned parameters."""
        # actors will be stored in derived classes
        params = {
                "critic_state_dicts": [cr.state_dict() for cr in self._critics]
                }
        return params


    def get_state(self):
        """Returns dictionary containing state."""
        state = {
                "critic_optim_state_dicts": [cr_optim.state_dict() for cr_optim in self._critic_optims]
                }
        return state


    def load_parameters(self, params):
        """Load all learned parameters from dictionary."""
        for critic, state_dict in zip(self._critics, params["critic_state_dicts"]):
            critic.load_state_dict(state_dict)


    def load_state(self, state):
        """Load state from dictionary."""
        for critic_optim, state_dict in zip(self._critic_optims, state["critic_optim_state_dicts"]):
            critic_optim.load_state_dict(state_dict)




import time
import csv
import json

class CSVLogger:

    def __init__(self, file_path, fieldnames, append = False):
        self.t_start = time.time()
        if append:
            self.file = open(file_path, "a")
        else:
            self.file = open(file_path, "w")
            self.file.write("#{}\n".format(json.dumps({"t_start": self.t_start})))
        self.csv_logger = csv.DictWriter(self.file, fieldnames = fieldnames)
        if not append: 
            self.csv_logger.writeheader()
        self.file.flush()

    def log(self, row_dict):
         self.csv_logger.writerow(row_dict)
         self.file.flush()

    def time_passed(self):
        return round(time.time() - self.t_start, 6)

def read_csv_log(file_path):
    columns = []
    column_dict = {}
    with open(file_path) as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=',')
        line_count = 0
        for row in csv_reader:
            if row[0][0] != "#":
                if line_count == 0:
                    for column_name in row:
                        column = []
                        column_dict[column_name] = column
                        columns.append(column)
                else:
                    for value, column in zip(row, columns):
                        column.append(float(value))
                line_count += 1
    return column_dict




import os
import torch

from ..nodes import Node
from ..data import FlatTransition

class Graph:
    """Graph defining the hierarchy.

    The attribute entry_node contains the root node of the graph while 
    active_node keeps track of which node is currently active.
    """

    def __init__(self, name, entry_node, env_action_space, global_algorithms = None):
        self.name = name
        self.entry_node = entry_node
        self.active_node = self.entry_node
        self.env_action_space = env_action_space
        self.global_algorithms = global_algorithms

        self._env_obs = None
        self._atomic_action = None
        self._env_return = 0
        self.tb_writer = None

        # create policies in the nodes recursively
        self.entry_node.create_policy(env_action_space)
        self.reset()

    def reset(self):
        self.active_node = self.entry_node
        self.active_parent = None
        self.active_parent_info = None

        self.entry_node.reset()

    def get_atomic_action(self, env_obs, sess_info, start_node = None, testing = False):
        self._env_obs = env_obs
        
        # forward pass through graph starting from active node/provided start node
        if start_node is not None:
            self.active_node = start_node

        assert self.active_node is not None

        action, self.active_node = self.active_node.get_action(
                env_obs = env_obs, 
                parent_info = self.active_parent_info, 
                parent = self.active_parent,
                sess_info = sess_info, 
                testing = testing)
        self._atomic_action = action

        return action

    def register_experience(self, env_info, sess_info):
        # First backward path starting from active node: check for interruption by parent
        parent_interruption, interrupting_node = self.active_node.check_parent_interruption(
                env_info, Node.sink_identification_feedback, sess_info)

        # Second backward pass through graph starting from active node:
        # Propagate feedback from children back to parents and register 
        # experience in nodes that acted.         
        self.active_node, self.active_parent, self.active_parent_info = \
                self.active_node.register_experience(env_info, Node.sink_identification_feedback, 
                        sess_info, interrupting_node)

        # global algorithms
        if self.global_algorithms:
            trans = FlatTransition(self._env_obs, self._atomic_action, env_info)
            for g_algo in self.global_algorithms:
                g_algo.add_experience(trans, sess_info)
                if learn:
                    g_algo.learn(sess_info)

        # is execution of graph finished?
        return self.active_node is None

    def learn(self, sess_info):
        """Learn without doing a rollout.

        Can be used for learning after instead of during a rollout."""
        nodes = self.entry_node.get_descendants_recursively()
        for node in nodes:
            node.algorithm.learn(sess_info.total_step)

    def set_tensorboard_writer(self, tb_writer):
        self._tb_writer = tb_writer
        self.entry_node.set_tensorboard_writer(tb_writer, [])

    def create_logfiles(self, logdir, append = False):
        nodes = self.entry_node.get_descendants_recursively()
        for node in nodes:
            node.create_logfiles(logdir, append)

    def get_logfiles(self):
        nodes = self.entry_node.get_children_recursively()
        logfile_dict = {}
        for node in nodes:
            logfile_dict.update(node.get_logfiles())
        return logfile_dict

    def get_node_ids(self):
        nodes = self.entry_node.get_descendants_recursively()
        node_id_dict = {}
        ID = 0
        for n in nodes:
            node_id_dict[n] = ID
            ID += 1
        return node_id_dict

    def get_parameters(self):
        """Returns dictionary containing all learned parameters in this graph."""
        return self.entry_node.get_parameters_recursively()

    def get_state(self):
        """Returns dictionary containing state of this graph.

        'State' refers to properties which influence training but not execution of the graph.
        This includes the state of the optimizers, replay buffers etc."""
        return self.entry_node.get_state_recursively()

    def load_parameters(self, params):
        """Load all learned parameters in this graph from file or dictionary."""
        if isinstance(params, str):
            params = torch.load(params)
        assert isinstance(params, dict)
        self.entry_node.load_parameters_recursively(params)

    def load_state(self, state = None, dir_path = None):
        """Load state of this graph from files in given directory or from dict."""
        assert {state, dir_path} != {None, None}, "Either state dict or directory with state files has to be specified."
        if state is not None:
            assert isinstance(state, dict)
            self.entry_node.load_state_recursively(state)
        else:
            s = torch.load(os.path.join(dir_path, "graph_state.pt"))
            self.entry_node.load_state_recursively(s, dir_path)

    def save_parameters(self, path):
        """Save all learned parameters in this graph to file."""
        params = self.get_parameters()
        torch.save(params, path)

    def save_state_pickle(self, path):
        """Save state of this graph to file using pickle/torch.save.
        
        'State' includes the state of the optimizers, replay buffers, 
        model parameters etc."""
        state = self.get_state()
        torch.save(state, path)

    def save_state(self, dir_path):
        """Save state of this graph to files in given directory.
        
        'State' includes the state of the optimizers, replay buffers, 
        model parameters etc."""
        node_ids = self.get_node_ids()
        state = self.entry_node.save_state_recursively(node_ids, dir_path)
        torch.save(state, os.path.join(dir_path, "graph_state.pt"))





import numpy as np
import gym


class DummyEnv(gym.GoalEnv):  
    """Env that only holds observation_space and action_space.

    Dummy environment can be used to communicate observation and action space 
    to stable baselines algorithms."""

    metadata = {'render.modes': ['human']}   

    def __init__(self, action_space, observation_space):
        self.action_space = action_space
        self.observation_space = observation_space

    def compute_reward(self, achieved_goal, desired_goal, info):
        pass

    def step(self, action):
        pass

    def reset(self):
        pass

    def render(self, mode='human', close=False):
        pass



from abc import ABC, abstractmethod

class Algorithm(ABC):
    """Algorithm base class."""

    def __init__(self, name):
        super().__init__()
        self.name = name
        self._tb_writer = None

    @abstractmethod
    def add_experience(self, env_obs, env_info, subtask_trans, parent_info, 
            child_feedback, algo_info, interruption, sess_info):
        pass

    @abstractmethod
    def learn(self, **kwargs):
        pass

    def get_algo_info(self, env_obs, parent_info):
        return None

    def set_tensorboard_writer(self, tb_writer):
        self._tb_writer = tb_writer

    @abstractmethod
    def get_parameters(self):
        """Returns dictionary containing all learned parameters."""
        pass

    @abstractmethod
    def get_state(self):
        """Returns dictionary containing state."""
        pass

    @abstractmethod
    def save_state(self, node_id):
        """Saves state and returns dictionary with file paths."""
        pass

    @abstractmethod
    def load_parameters(self, params):
        """Load all learned parameters from dictionary."""
        pass

    @abstractmethod
    def load_state(self, state, dir_path = None):
        """Load state from dictionary."""
        pass




from . import Node
from ..subtasks import TimedGoalSubtask, DictInfoHidingTolTGSubtaskSpec
from ..algorithms import HiTS

class HiTSNode(Node):

    def __init__(self, name, parents, subtask_spec, HiTS_kwargs, tb_writer = None):

        # create subtask
        subtask = TimedGoalSubtask(name + "_subtask", subtask_spec)

        # create algorithm
        algorithm = HiTS(
                name = name + "_algorithm",
                check_status = subtask.check_status, 
                convert_time = subtask.task_spec.convert_time, 
                delta_t_min = subtask.task_spec._delta_t_min,
                use_normal_trans_for_testing = False, 
                use_testing_transitions = True,  
                learn_from_deterministic_episodes = True, 
                **HiTS_kwargs)

        # policy creation is done via algorithm because it does the sampling part
        # of generating an action
        policy_class = algorithm.get_policy

        self.current_timed_goal = None
        self.current_goal_tol = None

        super(HiTSNode, self).__init__(name, policy_class, subtask, algorithm, parents)

    def check_parent_interruption(self, env_info, child_feedback, sess_info, 
            check_this_node = False):
        parent_info = self._parent_info[-1]
        self.current_timed_goal = self.subtask.get_updated_desired_tg_in_steps(env_info.new_obs, 
                parent_info, sess_info)
        if isinstance(self.subtask.task_spec, DictInfoHidingTolTGSubtaskSpec):
            self.current_goal_tol = parent_info.action["goal_tol"]
        elif hasattr(self.subtask.task_spec, "_goal_achievement_threshold"):
            self.current_goal_tol = self.subtask.task_spec._goal_achievement_threshold
        else:
            self.current_goal_tol = None
        return super(HiTSNode, self).check_parent_interruption(env_info, child_feedback, 
                sess_info, check_this_node)



from .env_info import EnvInfo
from .sess_info import SessInfo
from .parent_info import ParentInfo
from .flat_transition import FlatTransition
from .subtask_transition import SubtaskTransition
from .segmented_replay_buffer import SegmentedReplayBuffer



import torch
from torch import nn
import torch.nn.functional as F

class SquashNetwork(nn.Module):
    """Squashes output to n-dim interval via tanh."""
    def __init__(self, core_module, low, high, device = "cpu"):
        """
        Args: 
            core_module: Module that is used to process inputs before squashing.
        """
        super(SquashNetwork, self).__init__()

        self.core_module = core_module
        self.device = device
        self._offset = torch.tensor(0.5*(high + low), device = self.device, dtype = torch.float)
        self._scale = torch.tensor(0.5*(high - low), device = self.device, dtype = torch.float)


    def forward(self, x):
        if not isinstance(x, torch.Tensor):
            x = torch.tensor(x, device = self.device, dtype = torch.float)
        return self._scale*torch.tanh(self.core_module(x)) + self._offset



from . import AuxReward

class EnvReward(AuxReward):
    """Cumulative environment reward as auxiliary reward."""

    def __init__(self, weight):
        self._weight = weight
        self._cum_rew = 0.

    def __call__(self, obs, action, env_reward):
        aux_rew = self._weight*self._cum_rew
        self._cum_rew = 0
        return aux_rew

    def update(self, env_info):
        """Adds env reward to cumulative environment reward every time step."""

        self._cum_rew += env_info.reward



from .space import Space
from .box_space import BoxSpace
from .tuple_space import TupleSpace
from .multi_binary_space import MultiBinarySpace
from .dict_space import DictSpace
from .binary_space import BinarySpace

from .utils import space_from_gym_space



import os

from ..spaces import space_from_gym_space
from ..data import SubtaskTransition, ParentInfo

class Node:
    """Node in the directed graph defining the hierarchy.

    Contains a policy, a subtask and an algorithm."""

    sink_identification_feedback = object()

    def __init__(self, name, policy_class, subtask, algorithm, parents, 
            interruption_policy = None):
        self.name = name
        self._policy_class = policy_class
        self.policy = None
        self.subtask = subtask
        self.algorithm = algorithm
        self.tb_writer = None
        if interruption_policy is None:
            self.interruption_policy = lambda env_obs, subtask_obs: False
        else:
            self.interruption_policy = interruption_policy

        self._children = []
        self._parents = parents

        self._env_obs = []
        self._parent_info = []
        self._active_parent = []
        self._ind_child_to_activate = []
        self._action = []
        self._algo_info = []
        self._env_observation = []
        self._subtask_obs = []

        self._is_reset = True

    def reset(self):
        if self._is_reset == False:
            self._env_obs.clear()
            self._parent_info.clear()
            self._active_parent.clear()
            self._ind_child_to_activate.clear()
            self._action.clear()
            self._algo_info.clear()
            self._env_observation.clear()
            self._subtask_obs.clear()

            self.subtask.reset()

            self._is_reset = True

            for child in self._children:
                child.reset()

    def add_child(self, child):
        self._children.append(child)

    def add_parent(self, parent):
        self._parents.append(parent)

    def create_policy(self, env_action_space):
        # have to know all parents and children at this point
        if self.policy is None:
            p_name = self.name + "_policy"

            observation_space = self.subtask.observation_space

            if not self._children:
                action_space = space_from_gym_space(env_action_space)
            else:
                child_parent_action_space = self._children[0].subtask.parent_action_space
                assert all([child.subtask.parent_action_space == child_parent_action_space 
                    for child in self._children[1:]])
                action_space = child_parent_action_space

            self.policy = self._policy_class(p_name, observation_space, action_space)

            for child in self._children:
                child.create_policy(env_action_space)

    def get_action(self, env_obs, parent_info, parent, sess_info, testing = False):
        self._is_reset = False

        # use subtask to map environment observation and and parent information 
        # to observation for this node 
        subtask_obs = self.subtask.get_observation(env_obs, parent_info, sess_info)

        algo_info = self.algorithm.get_algo_info(env_obs, parent_info)

        ind_child_to_activate, action = self.policy(subtask_obs, algo_info, testing = testing)

        self._ind_child_to_activate.append(ind_child_to_activate)

        # keep track of env obs, active parent, parent info etc. in order to be able
        # register this experience for learning later on
        self._env_obs.append(env_obs)
        self._active_parent.append(parent)
        self._parent_info.append(parent_info)
        self._subtask_obs.append(subtask_obs)
        self._algo_info.append(algo_info)
        self._action.append(action)

        # if node is a sink (no children), return output of policy as 
        # (atomic) action and self as active node
        if not self._children:
            return action, self
        # otherwise get action from child that is to be activated
        else:
            parent_info_for_child = ParentInfo(action, algo_info, sess_info.ep_step)
            return self._children[ind_child_to_activate].get_action(
                    env_obs, parent_info_for_child, self, sess_info, testing)

    def check_parent_interruption(self, env_info, child_feedback, sess_info, check_this_node = False):
        if self._parents:
            # check whether parents want to interrupt (have priority over children)
            self._interrupted_by_parent, interrupting_parent_node = \
            self._active_parent[-1].check_parent_interruption(env_info, child_feedback, 
                    sess_info, check_this_node = True)
        else:
            self._interrupted_by_parent = False

        # check for interruption in this node
        if check_this_node:
            parent_info = self._parent_info[-1]
            new_subtask_obs = self.subtask.get_observation(env_info.new_obs, parent_info, sess_info)

            # check whether the interruption policy wants to interrupt
            self._interrupted_by_pol = self.interruption_policy(env_info.new_obs, new_subtask_obs)

            # check whether the subtask wants to interrupt
            self._interrupted_by_subtask = self.subtask.check_interruption(env_info, 
                    new_subtask_obs, parent_info, sess_info)
        else:
            self._interrupted_by_pol = False
            self._interrupted_by_subtask = False

        # interruption by parent has priority over interruption by child
        if self._interrupted_by_parent:
            return True, interrupting_parent_node
        elif self._interrupted_by_pol or self._interrupted_by_subtask:
            return True, self
        else:
            return False, None


    def register_experience(self, env_info, child_feedback, sess_info, interrupting_node = None):
        env_obs = self._env_obs.pop()
        ind_child_to_activate = self._ind_child_to_activate.pop()
        action = self._action.pop()
        algo_info = self._algo_info.pop()

        active_parent = self._active_parent.pop()
        parent_info = self._parent_info.pop()
        subtask_obs = self._subtask_obs.pop()

        new_subtask_obs = self.subtask.get_observation(env_info.new_obs, parent_info, sess_info)

        # incomplete subtask transition, still missing reward, ended and info
        st_trans_incmplt = SubtaskTransition(subtask_obs, action, ind_child_to_activate, new_obs = new_subtask_obs)

        subtask_trans, node_feedback = self.subtask.evaluate_transition(env_obs, env_info, 
                st_trans_incmplt, parent_info, algo_info, sess_info)

        if not sess_info.testing:
            self.algorithm.add_experience(env_obs, env_info, subtask_trans, parent_info, 
                    child_feedback, algo_info, interrupting_node is not None, sess_info)

        # TODO: Evaluate whether learning during an episode could be a 
        # useful feature
        #if sess_info.learn and not sess_info.testing:
        #    self.algorithm.learn(sess_info.total_step)

        # Return control to parent node if:
        # - there is an interrupting node and it is not this node
        # - the subtask in this node has ended
        # - the environment is done
        if (interrupting_node not in [None, self] or subtask_trans.ended or 
                env_info.done):
            # Return control to active parent if there is one
            if active_parent is not None:
                # current node can't interrupt parents
                if interrupting_node is self:
                    interrupting_node = None
                return active_parent.register_experience(env_info, node_feedback, 
                        sess_info, interrupting_node)
            # If there is no active parent return None in order to signal termination of graph
            else:
                return None, None, None
        # else return this node as new active node
        else:
            return self, active_parent, parent_info

    def get_descendants_recursively(self, visited = None):
        if visited is None:
            visited = set()
        visited.add(self)
        children = [self]
        for child in self._children:
            if child not in visited:
                children.extend(child.get_descendants_recursively(visited))
        return children

    def set_tensorboard_writer(self, tb_writer, visited_nodes):
        self._tb_writer = tb_writer
        self.subtask.set_tensorboard_writer(tb_writer)
        self.algorithm.set_tensorboard_writer(tb_writer)
        visited_nodes.append(self)
        for child in self._children:
            if child not in visited_nodes:
                child.set_tensorboard_writer(tb_writer, visited_nodes)
    
    def create_logfiles(self, logdir, append = False):
        self.subtask.create_logfiles(logdir, append)

    def get_logfiles(self):
        subtask_logfiles = self.subtask.get_logfiles()
        logfile_dict = {self.name: {
            "type": self.__class__.__name__, 
            "subtask_logfiles": self.subtask_logfiles
            } }
        return logfile_dict

    def get_parameters(self):
        """Returns dictionary containing all learned parameters of this node only."""
        return self.algorithm.get_parameters()

    def get_state(self):
        """Returns dictionary containing state of this node only."""
        return self.algorithm.get_state()

    def save_state(self, node_id, dir_path):
        """Saves state of this node only and returns dict with paths to files."""
        return self.algorithm.save_state(node_id, dir_path)

    def load_parameters(self, params):
        """Load all learned parameters from dictionary of this node only."""
        self.algorithm.load_parameters(params)

    def load_state(self, state, dir_path = None):
        """Load state of this node only."""
        self.algorithm.load_state(state, dir_path = dir_path)

    def get_parameters_recursively(self, visited = None):
        """Returns dictionary containing all learned parameters of this node and its descendants.
        
        The dictionary is assembled recursively so that the parameters of all nodes which 
        are reachable from this node are contained in it."""
        if visited is None:
            visited = set()
        visited.add(self)
        children_params = []
        for child in self._children:
            if child in visited:
                child_params = None
            else:
                child_params = child.get_parameters_recursively(visited)
            children_params.append(child_params)
        params = {
                "node_params": self.get_parameters(),
                "children_params": children_params
                }
        return params

    def get_state_recursively(self, visited = None):
        """Returns dictionary containing the state of this node and its descendants.
        
        The dictionary is assembled recursively so that the state of all nodes which 
        are reachable from this node are contained in it."""
        if visited is None:
            visited = set()
        visited.add(self)
        children_state = []
        for child in self._children:
            if child in visited:
                child_state = None
            else:
                child_state = child.get_state_recursively(visited)
            children_state.append(child_state)
        state = {
                "node_state": self.get_state(),
                "children_state": children_state
                }
        return state

    def save_state_recursively(self, node_ids, dir_path, visited = None):
        """Returns dictionary containing the state of this node and its descendants.
        
        Note: In this version the state may contain paths to auxiliary files that have 
        to be taken into account when loading the state. The dictionary is assembled 
        recursively so that the state of all nodes which are reachable from this node 
        are contained in it."""
        if visited is None:
            visited = set()
        visited.add(self)
        children_state = []
        for child in self._children:
            if child in visited:
                child_state = None
            else:
                child_state = child.save_state_recursively(node_ids, dir_path, visited)
            children_state.append(child_state)
        state = {
                "node_state": self.save_state(node_ids[self], dir_path),
                "children_state": children_state
                }
        return state


    def load_parameters_recursively(self, params):
        """Load all learned parameters of this node and all nodes reachable from it."""
        assert len(self._children) == len(params["children_params"])
        for child, child_params in zip(self._children, params["children_params"]):
            if child_params is not None:
                child.load_parameters_recursively(child_params)
        self.load_parameters(params["node_params"])

    def load_state_recursively(self, state, dir_path = None):
        """Load state of this node and all nodes reachable from it."""
        assert len(self._children) == len(state["children_state"])
        for child, child_state in zip(self._children, state["children_state"]):
            if child_state is not None:
                child.load_state_recursively(child_state, dir_path)
        self.load_state(state["node_state"], dir_path)



import numpy as np

from . import Graph
from ..nodes import HiTSNode, HACNode
from ..utils import listify
from ..aux_rewards import EnvReward

class HiTSGraph(Graph):

    def __init__(self, name, n_layers, env, subtask_specs, HAC_kwargs, HiTS_kwargs, 
            update_sgs_rendering=None, update_tsgs_rendering=None, env_reward_weight=None):
        """
        Args:
        n_layers (int): Number of layers in hierarchy.
        """

        self.n_layers = n_layers
        self._update_sgs_rendering = update_sgs_rendering 
        self._update_tsgs_rendering = update_tsgs_rendering 

        # convert parameters to list with one entry per layer if not already a list
        HiTS_kwargs_list = listify(HiTS_kwargs, n_layers - 1)
        subtask_specs = listify(subtask_specs, n_layers - 1)

        # by default penalty for choosing a goal which was not reached by the child is minus
        # maximum number of actions in episode on this layer
        if "child_failure_penalty" not in HAC_kwargs:
            HAC_kwargs["child_failure_penalty"] = -subtask_specs[-1].max_n_actions
        for kwargs, subtask_spec in zip(HiTS_kwargs_list, subtask_specs):
            if "child_failure_penalty" not in kwargs:
                kwargs["child_failure_penalty"] = -1.

        # let higest level see environment reward in addition to penalty
        # for emitting timed subgoals if desired
        if env_reward_weight is not None:
            subtask_specs[-1].add_aux_reward(EnvReward(env_reward_weight))

        # highest node is HAC node because most environments do not provide a 
        # timed subgoal
        highest_node = HACNode(
                    name = "hac_node_layer_{}".format(n_layers - 1), 
                    parents = [],
                    subtask_spec = subtask_specs[-1], 
                    HAC_kwargs = HAC_kwargs)

        # construct graph
        parents = [highest_node]
        self._nodes = [highest_node]
        for l in range(n_layers - 2, -1, -1):
            node_name = "hits_node_layer_{}".format(l)
            node = HiTSNode(
                    name = node_name, 
                    parents = parents,
                    subtask_spec = subtask_specs[l], 
                    HiTS_kwargs = HiTS_kwargs_list[l])
            if l < n_layers - 1:
                self._nodes[0].add_child(node)
            self._nodes.insert(0, node)
            parents = [node]

        entry_node = self._nodes[-1]


        super(HiTSGraph, self).__init__(name, entry_node, env.action_space)

    def get_atomic_action(self, env_obs, sess_info, start_node = None, testing = False):
        action = super().get_atomic_action(env_obs, sess_info, start_node, testing)

        # update timed subgoals via callback for rendering them in environment
        if self._update_tsgs_rendering is not None:
            subgoals = [nd.current_timed_goal for nd in self._nodes[:-1]]
            tolerances = [nd.current_goal_tol for nd in self._nodes[:-1]]
            self._update_tsgs_rendering(subgoals, tolerances)

        return action




from .aux_reward import AuxReward
from .goal_tol_reward import GoalTolReward
from .power_reward import PowerReward
from .constant_reward import ConstantReward
from .delta_t_ach_reward import DeltaTAchReward
from .env_reward import EnvReward



from abc import ABC, abstractmethod

import numpy as np

from ..spaces import BoxSpace, DictSpace, space_from_gym_space
from ..utils import listify, get_obs_space_from_env, get_obs_from_gym
from .subtask_specs import SubtaskSpec

# TODO: Refactor inheritance. Also compare to TimedGoalSubtaskSpecs.

class ShortestPathSubtaskSpec(SubtaskSpec):
    """Specification of shortest path subtask, i.e. criterion of goal achievement etc."""

    def __init__(self, constant_failure_return = False):
        super().__init__()
        self.constant_failure_return = constant_failure_return 

    @abstractmethod
    def goal_achievement_criterion(self, achieved_goal, desired_goal, parent_info):
        """Maps achieved_goal, desired_goal and parent_info to True/False depending on whether desired_goal was achieved.
        
        Note that parent_info can contain information about the tolerance for achieving a subgoal."""
        pass

    @abstractmethod
    def get_desired_goal(self, env_obs, parent_info):
        """Maps env_obs and parent_info to desired goal in goal_space."""
        pass

    @abstractmethod
    def map_to_goal(self, partial_obs):
        """Maps partial_obs to goal (assumed to be a deterministic map with image in goal_space)."""
        pass

    @abstractmethod
    def map_to_partial_obs(self, env_obs, parent_info, ep_step):
            """Maps env_obs and parent_info to partial observation (assumed to be a deterministic map with image in partial_obs_space)."""
            pass

    @property
    @abstractmethod
    def partial_obs_space(self):
        """Box space for partial observations (partial observation does not contain desired goal)."""
        pass

    @property
    @abstractmethod
    def goal_space(self):
        """Box space containing goals of the subtask."""
        pass

    @property
    @abstractmethod
    def parent_action_space(self):
        """Space which contains actions the parent node is taking. 
        
        The action of a parent determines the the goal of the subtask via the method get_desired_goal."""
        pass

    @property
    @abstractmethod
    def max_n_actions(self):
        """Maximum number of actions before subtask ends."""
        pass

class BoxSPSubtaskSpec(ShortestPathSubtaskSpec):
    """Default subtask spec with goal space being equal to state box space.

    The goal is equal to the parent action, partial observation is partial env 
    observation and achievement criterion uses L2 norm.
    
    If factorization is not None, the space is split up and the L2 norm is 
    applied to each subspace individually (with different thresholds). This 
    can be useful if the goal contains different types of values, e.g., 
    positions and derivatives."""
    def __init__(self, max_n_actions, goal_achievement_threshold, env, factorization = None, 
            constant_failure_return = False):
        super().__init__(constant_failure_return)
        self._max_n_actions = max_n_actions
        # Observation and goal spaces are all identical in this case
        self._space = get_obs_space_from_env(env)
        if factorization is None:
            self.factorization = [range(self._space.get_flat_space().n)]
        else:
            self.factorization = factorization
        self._goal_achievement_threshold = listify(goal_achievement_threshold, 1)
        assert len(self.factorization) == len(self._goal_achievement_threshold)

    def goal_achievement_criterion(self, achieved_goal, desired_goal, parent_info):
        # if the criterion in one of the subspaces is violated, the overall criterion is not met
        for factor, threshold in zip(self.factorization, self._goal_achievement_threshold):
            if np.linalg.norm(achieved_goal[factor] - desired_goal[factor]) >= threshold:
                return False
        return True

    def get_desired_goal(self, env_obs, parent_info):
        # desired goal is equal to parent action, no processing required
        return parent_info.action

    def map_to_goal(self, partial_obs):
        # achieved goal is identical to partial observation
        return partial_obs

    def map_to_partial_obs(self, env_obs, parent_info, ep_step):
        # all levels see full partial observation (but not env goal)
        return get_obs_from_gym(env_obs)

    @property
    def partial_obs_space(self):
        return self._space

    @property
    def goal_space(self):
        return self._space

    @property
    def parent_action_space(self):
        return self._space

    @property
    def max_n_actions(self):
        return self._max_n_actions

class BoxInfoHidingSPSubtaskSpec(BoxSPSubtaskSpec):
    """Subtask spec with goal space composed of only some of the dimensions of the
    state space.

    The goal is equal to the parent action, the partial observation is obtained 
    by picking a subset of the components of the env partial observation.  
    The achieved goal is obtained from the partial obs by concatenating the 
    dimensions in goal_indices. See DefaultSPSubtaskSpec for more details."""
    def __init__(self, max_n_actions, goal_achievement_threshold, partial_obs_indices, 
            goal_indices, env, factorization = None, constant_failure_return = False):
        super(BoxSPSubtaskSpec, self).__init__(constant_failure_return)
        if factorization is None:
            factorization = [range(len(goal_indices))]
        super(BoxInfoHidingSPSubtaskSpec, self).__init__(max_n_actions, 
                goal_achievement_threshold, env, factorization)
        self._partial_obs_indices = partial_obs_indices
        self._goal_indices = range(len(partial_obs_indices)) if goal_indices is None else goal_indices

        self._space = self._space.get_flat_space()
        self._space = self._space.get_flat_space(indices = partial_obs_indices)
        self._goal_space = self._space.get_flat_space(indices = goal_indices)

    def map_to_goal(self, partial_obs):
        # achieved goal is constructed from some of the dimensions of partial obs
        return partial_obs[self._goal_indices]

    def map_to_partial_obs(self, env_obs, parent_info, ep_step):
        # partial observation is constructed by concatenating subset of 
        # components of partial env obs
        return get_obs_from_gym(env_obs)[self._partial_obs_indices]

    @property
    def goal_space(self):
        return self._goal_space

    @property
    def parent_action_space(self):
        return self._goal_space

class DictSPSubtaskSpec(BoxSPSubtaskSpec):
    """Subtask spec with goal space being equal to state dict space.

    The goal is equal to the parent action, partial observation is partial env 
    observation and achievement criterion uses L2 norm.

    The goal space is assumed to be a dict space and the L2 norm in the subspaces 
    is used together with the values in the goal_achievement_threshold dict as 
    thresholds.  This can be useful if the goal contains different types of values, 
    e.g., positions and derivatives."""
    
    def __init__(self, max_n_actions, goal_achievement_threshold, env, 
            constant_failure_return = False):
        super(BoxSPSubtaskSpec, self).__init__(constant_failure_return)
        assert isinstance(goal_achievement_threshold, dict)
        self._max_n_actions = max_n_actions
        # Partial observation, parent action and goal spaces are all identical in this case
        self._space = get_obs_space_from_env(env)
        self._goal_achievement_threshold = goal_achievement_threshold

    def goal_achievement_criterion(self, achieved_goal, desired_goal, parent_info):
        # if the criterion in one of the subspaces is violated, the overall criterion is not met
        for key, threshold in self._goal_achievement_threshold.items():
            if np.linalg.norm(achieved_goal[key] - desired_goal[key]) >= threshold:
                return False
        return True


class DictInfoHidingSPSubtaskSpec(DictSPSubtaskSpec):
    """Subtask spec with partial observation and goal dict space composed of 
    only some of the items of the env observation dict space.

    The desired goal is equal to the parent action. The partial observation is 
    obtained by picking a subset of the keys of the env partial observation. 
    The achieved goal is obtained from the partial obs by concatenating the 
    items with keys in goal_keys. See DefaultSPSubtaskSpec for more details."""
    def __init__(self, max_n_actions, goal_achievement_threshold, partial_obs_keys, 
            goal_keys, env, constant_failure_return = False, max_ep_steps = 1.):
        super(DictInfoHidingSPSubtaskSpec, self).__init__(max_n_actions, 
                goal_achievement_threshold, env, constant_failure_return)
        self._partial_obs_keys = partial_obs_keys
        self._goal_keys = goal_keys

        self._space = self._space.get_subspace(partial_obs_keys)
        if "__ep_time__" in self._partial_obs_keys:
            # add time item to dict space
            new_dict_space = {
                **self._space._space_dict, 
                "__ep_time__": BoxSpace([0.], [1.])
            }
            self._space = DictSpace(new_dict_space)
        self._goal_space = self._space.get_subspace(goal_keys)
        self._max_ep_steps = max_ep_steps

    def map_to_goal(self, partial_obs):
        return {key: value for key, value in partial_obs.items() if key in self._goal_keys}

    def map_to_partial_obs(self, env_obs, parent_info, ep_step):
        p_obs = {key: value for key, value in get_obs_from_gym(env_obs).items() if key in self._partial_obs_keys}
        if "__ep_time__" in self._partial_obs_keys:
            p_obs["__ep_time__"] = np.clip([ep_step/self._max_ep_steps], 0., 1.)
        return p_obs

    @property
    def goal_space(self):
        return self._goal_space

    @property
    def parent_action_space(self):
        return self._goal_space

# subtask specification based on goal based environment
class EnvSPSubtaskSpec(ShortestPathSubtaskSpec):
    def __init__(self, max_n_actions, env, map_to_env_goal, 
            constant_failure_return=False, achievement_reward=0.):
        super().__init__(constant_failure_return)
        self._max_n_actions = max_n_actions
        self._partial_env_obs_space = space_from_gym_space(env.observation_space["observation"])
        self._env_goal_space = space_from_gym_space(env.observation_space["desired_goal"])
        self._map_to_env_goal = map_to_env_goal
        self._env = env
        self._achievement_reward = achievement_reward

    def goal_achievement_criterion(self, achieved_goal, desired_goal, parent_info):
        reward = self._env.compute_reward(achieved_goal, desired_goal, {})
        return reward == self._achievement_reward

    def get_desired_goal(self, env_obs, parent_info):
        return env_obs["desired_goal"]

    def map_to_goal(self, partial_obs):
        return self._map_to_env_goal(partial_obs)

    def map_to_partial_obs(self, env_obs, parent_info, ep_step):
        # highest level sees environment
        return env_obs["observation"]

    @property
    def partial_obs_space(self):
        return self._partial_env_obs_space

    @property
    def goal_space(self):
        return self._env_goal_space

    @property
    def parent_action_space(self):
        return None

    @property
    def max_n_actions(self):
        return self._max_n_actions



import numpy as np
import pyglet.gl as gl

from .graphics_utils import ArrowConfig

def draw_circle_sector(center, angle, radius, n, color, triangles_to_draw):
    gl.glBegin(gl.GL_TRIANGLE_FAN)
    gl.glColor3f(*color)
    gl.glVertex2f(*center)
    for i in range(triangles_to_draw + 1):
        gl.glVertex3f(center[0] + np.cos(2.*np.pi/n*i + angle)*radius, 
                      center[1] + np.sin(2.*np.pi/n*i + angle)*radius,
                      0.0)
    gl.glEnd()

def draw_box(center, diameter, length, phi, color):
    gl.glPushMatrix()
    gl.glLoadIdentity()
    gl.glTranslatef(center[0], center[1], 0.)
    gl.glRotatef(phi, 0., 0., 1.)
    gl.glBegin(gl.GL_QUADS)
    gl.glColor3f(*color)
    gl.glVertex2f(-0.5*diameter, -0.5*length)
    gl.glVertex2f(0.5*diameter, -0.5*length)
    gl.glVertex2f(0.5*diameter, 0.5*length)
    gl.glVertex2f(-0.5*diameter, 0.5*length)
    gl.glEnd()
    gl.glPopMatrix()

def draw_line(center, radius, angle, line_color):
    gl.glBegin(gl.GL_LINES)
    gl.glColor3f(*line_color)
    gl.glVertex2f(center[0] - np.cos(angle)*radius, 
    center[1] - np.sin(angle)*radius)
    gl.glVertex2f(center[0] - np.cos(angle + np.pi)*radius, 
    center[1] - np.sin(angle + np.pi)*radius)
    gl.glEnd()


def draw_vector(initial_point, vector, arrow_config):
    v = vector*arrow_config.scale
    length = np.linalg.norm(v)
    width = arrow_config.width
    arrow_head_size = arrow_config.head_size
    color = arrow_config.color
    # orthogonal vector used for constructing vertices that make up 
    # arrow shape
    w = np.array([v[1]/length, -v[0]/length])
    # factor scaling vector such that it only reaches to arrow head and not to tip
    rescale = (length - arrow_config.head_size)/length
    v_re = np.array(v)*rescale

    gl.glPushMatrix()
    gl.glLoadIdentity()
    gl.glTranslatef(initial_point[0], initial_point[1], 0.)
    gl.glBegin(gl.GL_TRIANGLES)
    gl.glColor3f(*color)
    # rectangle
    gl.glVertex2f( 0.5*w[0]*width,              0.5*w[1]*width)
    gl.glVertex2f( 0.5*w[0]*width + v_re[0],  0.5*w[1]*width + v_re[1])
    gl.glVertex2f(-0.5*w[0]*width + v_re[0], -0.5*w[1]*width + v_re[1])
    gl.glVertex2f(-0.5*w[0]*width + v_re[0], -0.5*w[1]*width + v_re[1])
    gl.glVertex2f(-0.5*w[0]*width,             -0.5*w[1]*width)
    gl.glVertex2f( 0.5*w[0]*width,              0.5*w[1]*width)
    # arrow head
    gl.glVertex2f( 0.5*w[0]*arrow_head_size + v_re[0],  
            0.5*w[1]*arrow_head_size + v_re[1])
    gl.glVertex2f(-0.5*w[0]*arrow_head_size + v_re[0], 
            -0.5*w[1]*arrow_head_size + v_re[1])
    gl.glVertex2f(v[0], v[1])
    gl.glEnd()
    gl.glPopMatrix()

def draw_vector_with_outline(initial_point, vector, arrow_config, outline_color, inner_shrink_factor = 0.6):
    ac_outline = ArrowConfig(arrow_config.scale, arrow_config.width, arrow_config.head_size, outline_color)
    length = np.linalg.norm(vector) 
    scale_inner = (arrow_config.scale*length - arrow_config.head_size*(1. - inner_shrink_factor))/length
    ac_inner = ArrowConfig(scale_inner, arrow_config.width*inner_shrink_factor, 
        arrow_config.head_size*inner_shrink_factor, arrow_config.color)
    draw_vector(initial_point, vector, ac_outline)
    draw_vector(initial_point, vector, ac_inner)

def draw_circular_subgoal(position, velocity, radius, color, arrow_config, n_triangles = 32):
    draw_circle_sector(position, 
            0., 
            radius,
            n_triangles, 
            (0., 0., 0.),
            n_triangles)
    draw_circle_sector(position, 
            0., 
            0.8*radius,
            n_triangles, 
            color,
            n_triangles)
    if velocity is not None:
        draw_vector_with_outline(position, velocity, arrow_config, (0., 0., 0.))



from abc import ABC, abstractmethod
from copy import copy
import os

import tianshou as ts
import numpy as np
import torch

from . import Algorithm
from ..nodes import Node
from ..policies import TianshouPolicy

class OffPolicyAlgorithm(Algorithm):
    """Off-policy algorithm. Supports SAC and DDPG."""

    supported_flat_offpol_algos = {
            "SAC": ts.policy.SACPolicy,
            "DDPG": ts.policy.DDPGPolicy
            }

    class _Transition():
        def __init__(self, subtask_tr, env_info, child_feedback, algo_info):
            self.subtask_tr = subtask_tr
            self.env_info = env_info
            self.child_feedback = child_feedback
            self.algo_info = algo_info

        def __str__(self):
            return "subtask_tr: {}\n".format(self.subtask_tr) + \
            "child_feedback: {}\n".format(self.child_feedback) + \
            "algo_info: {}\n".format(self.algo_info)

    def __init__(self, name, flat_algo_name, model, fully_random_fraction, flat_algo_kwargs=None, 
            buffer_size=100000, batch_size=128, learning_starts=0, grad_steps_per_env_step=1.0):
        """
        Args:
            model (Model): Model instance which provides...
            for SAC:
            * actor: Pytorch module that returns ((mu, sigma), hidden_state) in forward method. mu and sigma refer to value in R^n before squashing.
            * two critics
            for DDPG:
            * actor: Pytorch module that returns (action, hidden_state) in forward method. action refers to value in R^n before squashing.
            * one critic

            learning_starts: Learnings starts after this many transitions have been collected.
        """
        super().__init__(name)

        if flat_algo_name not in self.supported_flat_offpol_algos:
            raise ValueError("Flat off policy RL algorithm \"{}\" not supported.".
                    format(flat_algo_name))

        assert model is not None

        self._flat_algo_name = flat_algo_name
        self._flat_algo_class = self.supported_flat_offpol_algos[flat_algo_name]
        self._flat_algorithm = None
        if flat_algo_kwargs is None:
            self._flat_algo_kwargs = {}
        else:
            self._flat_algo_kwargs = copy(flat_algo_kwargs)
        self._model = model
        self._fully_random_fraction = fully_random_fraction
        self._tb_low_level_writer = None
        self._actions_taken = 0
        self._buffer_size = buffer_size
        self._batch_size = batch_size
        self._learning_starts = learning_starts
        self._grad_steps_per_env_step = grad_steps_per_env_step
        self._episode_transitions = []

        # if a target entropy is given and no optimizer for the entropy coefficient is 
        # specified use adam and interpret third entry of tuple as learning rate
        if flat_algo_name == "SAC" and "alpha" in self._flat_algo_kwargs: 
            if (isinstance(self._flat_algo_kwargs["alpha"], list) 
                and not isinstance(self._flat_algo_kwargs["alpha"][0], torch.optim.Optimizer)):
                target_entropy = self._flat_algo_kwargs["alpha"][0]
                log_alpha = torch.tensor([self._flat_algo_kwargs["alpha"][1]], dtype = torch.float, requires_grad = True)
                alpha_optim = torch.optim.Adam([log_alpha], lr = self._flat_algo_kwargs["alpha"][2])
                self._flat_algo_kwargs["alpha"] = (target_entropy, log_alpha, alpha_optim)


        # set action range to [-1., 1.] in tianshou and rescale in graph_rl if needed
        self._flat_algo_kwargs["action_range"] = [-1., 1.]

        self._last_learning_step = None

        self._replay_buffer = ts.data.ReplayBuffer(size = buffer_size)

    def _map_action_to_default_space(self, action):
        res = (action - self._action_offset)/self._action_scale
        res = np.clip(res, -1., 1.)
        return res

    def _add_to_flat_replay_buffer(self, f_trans):
        # map action back to [-1, 1]^n before adding transition to buffer
        action = self._action_space.flatten_value(f_trans.action)
        mapped_action = self._map_action_to_default_space(action)
        self._replay_buffer.add(
                obs = self._observation_space.flatten_value(f_trans.obs), 
                act = mapped_action, 
                rew = f_trans.reward, 
                done = f_trans.done,
                obs_next = self._observation_space.flatten_value(f_trans.new_obs))

    def get_policy(self, name, observation_space, action_space):
        self._observation_space = observation_space
        self._action_space = action_space
        flat_action_space = action_space.get_flat_space()
        self._action_scale = 0.5*(flat_action_space.high - flat_action_space.low)
        self._action_offset = 0.5*(flat_action_space.high + flat_action_space.low)
        flat_obs_dim = observation_space.get_flat_dim()
        flat_action_dim = action_space.get_flat_dim()
        flat_gym_action_space = flat_action_space.get_gym_space()

        self._model.create(flat_obs_dim, flat_action_dim)
        self._learning_rate_critics = self._model._learning_rate_critics
        self._learning_rate_actor = self._model._learning_rate_actor

        # actor and critic modules
        nets = {
                "actor": self._model.actor,
                "actor_optim": self._model.actor_optim, 
                f"critic{'1' if self._flat_algo_name == 'SAC' else ''}": self._model.critics[0], 
                f"critic{'1' if self._flat_algo_name == 'SAC' else ''}_optim": self._model.critic_optims[0]
                }
        if self._flat_algo_name == "SAC":
            # SAC expects two critics while DDPG wants only one
            nets["critic2"] = self._model.critics[1]
            nets["critic2_optim"] = self._model.critic_optims[1]
        elif self._flat_algo_name == "DDPG":
            # Gaussian noise with n-dim. mean and variance
            class GaussianNoiseND(ts.exploration.random.BaseNoise):
                """Gaussian noise with n-dim. mean and variance."""

                def __init__(self, mu=[0.0], sigma=[1.0]):
                    super().__init__()
                    self._mu = mu
                    assert np.all(0 <= sigma), "Noise std should not be negative."
                    self._sigma = sigma

                def __call__(self, size):
                    # Note: size has to coincide with shape of mu and sigma
                    return np.random.normal(self._mu, self._sigma)

            # DDPG needs a noise object
            # flatten noise sigma object to vector using action space
            sigma = action_space.flatten_value(self._flat_algo_kwargs["noise_sigma"])
            exploration_noise = GaussianNoiseND(mu=np.zeros_like(sigma), sigma=sigma)
            self._flat_algo_kwargs["exploration_noise"] = exploration_noise
            del self._flat_algo_kwargs["noise_sigma"]

        self._flat_algorithm = self._flat_algo_class(**nets, **self._flat_algo_kwargs, 
                action_space=flat_gym_action_space)
        self._flat_algorithm.eval()

        policy = TianshouPolicy(
                name = name, 
                observation_space = observation_space, 
                action_space = action_space, 
                ts_policy = self._flat_algorithm, 
                hidden_net = self._model.actor,
                fully_random_fraction = self._fully_random_fraction)

        return policy

    @abstractmethod
    def _add_experience_to_flat_algo(self, parent_info, deterministic_episode, node_is_sink, sess_info):
        """Process transitions from ending episode and store them in flat replay buffer.

        At this point hindsight manipulations can be implemented.
        """

        pass

    def add_experience(self, env_obs, env_info, subtask_trans, parent_info, 
            child_feedback, algo_info, interruption, sess_info):

        # save transition to episode transitions buffer
        transition = self._Transition(
                subtask_tr = subtask_trans,
                env_info = env_info,
                child_feedback = child_feedback,
                algo_info = algo_info
                )
        self._episode_transitions.append(transition)
        if self._verbose:
            print(transition)
        self._actions_taken += 1

        # tensorboard logging
        if self._tb_writer is not None:
            obs = self._observation_space.flatten_value(subtask_trans.obs)
            if self._log_q_values and len(self._episode_transitions) == 1:
                act = self._action_space.flatten_value(subtask_trans.action)
                with torch.no_grad():
                    q_values = [critic(obs[None], act[None]).numpy() for critic  in self._model.critics]
                    q_value = min(q_values)
                mode = "test" if sess_info.testing else "train"
                self._tb_writer.add_scalar(f"{self.name}/{mode}/q_value", q_value, sess_info.total_step)
            if self._flat_algo_name == "SAC" and not sess_info.testing and parent_info is None:
                with torch.no_grad():
                    net_out, _ = self._model.actor(obs[None])
                    _, sigma = net_out
                    entropy = np.log(sigma.numpy()).sum() + (0.5 + np.log(np.sqrt(2.*np.pi)))*len(sigma)
                self._tb_writer.add_scalar(f"{self.name}/train/entropy", entropy, sess_info.total_step)

        # if the subtask has ended (i.e. control will be returned to active parent
        # node), process episode transitions and add them to replay buffer 
        # of flat algorithm
        if subtask_trans.ended or env_info.done:
            if parent_info:
                deterministic_episode = algo_info["is_deterministic"]
            else:
                deterministic_episode = False

            node_is_sink = child_feedback is Node.sink_identification_feedback
            self._add_experience_to_flat_algo(parent_info, deterministic_episode, 
                    node_is_sink, sess_info)

    def learn(self, current_step):
        if self._last_learning_step is None:
            self._last_learning_step = current_step - 1

        elapsed_steps = current_step - self._last_learning_step
        n_grad_steps = int(self._grad_steps_per_env_step*elapsed_steps)
        assert(elapsed_steps > 0)
        if len(self._replay_buffer) >= self._batch_size and current_step > self._learning_starts:
            for _ in range(int(n_grad_steps)):
                batch, indices = self._replay_buffer.sample(self._batch_size)
                batch = self._flat_algorithm.process_fn(batch, self._replay_buffer, indices)
                results = self._flat_algorithm.learn(batch)

                # tensorboard logging of losses
                if self._tb_writer is not None:
                    for key, item in results.items():
                        self._tb_writer.add_scalar(f"{self.name}/train/{key}", item, current_step)

        self._last_learning_step = current_step

    def get_parameters(self):
        """Returns dictionary containing all learned parameters."""
        params = {
                "flat_algo_name": self._flat_algo_name, 
                "model_params": self._model.get_parameters()
                }
        return params

    def get_state(self):
        """Returns dictionary containing state."""
        state = {
                "flat_algo_name": self._flat_algo_name, 
                "model_state": self._model.get_state(), 
                "model_params": self._model.get_parameters(), 
                "replay_buffer": self._replay_buffer
                }
        return state

    def save_state(self, node_id, dir_path):
        """Saves state and returns dictionary with file paths."""
        replay_buffer_file_name = f"{node_id}_algo_replay_buffer.hdf5"
        self._replay_buffer.save_hdf5(os.path.join(dir_path, replay_buffer_file_name))
        state = {
                "flat_algo_name": self._flat_algo_name, 
                "model_state": self._model.get_state(), 
                "model_params": self._model.get_parameters(), 
                "replay_buffer": replay_buffer_file_name
                }
        return state

    def _override_target_networks(self):
        """Override target networks with current networks."""
        tau = self._flat_algorithm._tau
        self._flat_algorithm._tau = 1.0
        self._flat_algorithm.sync_weight()
        self._flat_algorithm._tau = tau

    def load_parameters(self, params):
        """Load all learned parameters from dictionary."""
        assert self._flat_algo_name == params["flat_algo_name"]
        self._model.load_parameters(params["model_params"])
        self._override_target_networks()

    def load_state(self, state, dir_path = None):
        """Load state from dictionary."""
        assert self._flat_algo_name == state["flat_algo_name"]
        self._model.load_state(state["model_state"])
        self._model.load_parameters(state["model_params"])
        self._override_target_networks()
        if isinstance(state["replay_buffer"], ts.data.ReplayBuffer):
            self._replay_buffer = state["replay_buffer"]
        else:
            del self._replay_buffer
            self._replay_buffer = ts.data.ReplayBuffer.load_hdf5(
                    os.path.join(dir_path, state["replay_buffer"])
                    )



class FlatTransition():
    def __init__(self, obs, action, reward, new_obs, done):
        self.obs = obs
        self.action = action
        self.reward = reward
        self.new_obs = new_obs
        self.done = done

    @classmethod
    def from_env_info(cls, obs, action, env_info):
        return cls(obs, action, env_info.reward, env_info.new_obs, env_info.done)

    def __str__(self):
        return "obs: {}\n".format(self.obs) + \
        "action: {}\n".format(self.action) + \
        "reward: {}\n".format(self.reward) + \
        "new_obs: {}\n".format(self.new_obs) + \
        "done: {}\n".format(self.done)    




import torch

from .mlp_actor_critic_model import MLPActorCriticModel
from ..networks import DDPGNetwork

class DDPGModel(MLPActorCriticModel):
    def __init__(self, hidden_layers_actor, hidden_layers_critics, activation_fns_actor, 
            activation_fns_critics, learning_rate_actor, learning_rate_critics, 
            device, squash_critics = False, low = None, high = None, 
            critic_clip_threshold = None, actor_clip_threshold = None, force_negative = False):
        super(DDPGModel, self).__init__(hidden_layers_actor, hidden_layers_critics, 
                activation_fns_actor, activation_fns_critics, 1, learning_rate_actor, 
                learning_rate_critics, device, force_negative = force_negative)


    def create(self, state_dim, action_dim):
        super(DDPGModel, self).create(state_dim, action_dim)
        self._actor = DDPGNetwork(self._actor_stump, 1., self._hidden_layers_actor[-1], 
                action_dim, device = self._device)
        self._actor_optim = torch.optim.Adam(self._actor.parameters(), 
                lr = self._learning_rate_actor)

    @property
    def actor(self):
        return self._actor

    @property
    def actor_optim(self):
        return self._actor_optim

    def get_parameters(self):
        """Returns dictionary containing all learned parameters."""
        params = {
                "actor_state_dict": self._actor.state_dict(), 
                "mlp_params": super(DDPGModel, self).get_parameters()
                }
        return params

    def get_state(self):
        """Returns dictionary containing state."""
        state = {
                "actor_optim_state_dict": self._actor_optim.state_dict(), 
                "mlp_state": super(DDPGModel, self).get_state()
                }
        return state

    def load_parameters(self, params):
        """Load all learned parameters from dictionary."""
        super(DDPGModel, self).load_parameters(params["mlp_params"])
        self._actor.load_state_dict(params["actor_state_dict"])

    def load_state(self, state):
        """Load state from dictionary."""
        super(DDPGModel, self).load_state(state["mlp_state"])
        self._actor_optim.load_state_dict(state["actor_optim_state_dict"])



import numpy as np
import torch

from tianshou.data import Batch

from . import Policy
from ..networks import SACNetwork, DDPGNetwork

class TianshouPolicy(Policy):

    def __init__(self, name, observation_space, action_space, ts_policy, hidden_net = None, fully_random_fraction = 0.):
        """
        Args: 
            ts_policy: Tianshou policy. 
            hidden_net: Pytorch module returning either (mu, sigma), hidden_state for SAC or just action, hidden_state for DDPG.
        """
        super(TianshouPolicy, self).__init__(name, observation_space, action_space)
        self._fully_random_fraction = fully_random_fraction
        self._ts_policy = ts_policy
        self._hidden_net = hidden_net
        flat_action_space = self.action_space.get_flat_space()
        self._action_scale = 0.5*(flat_action_space.high - flat_action_space.low)
        self._action_bias = 0.5*(flat_action_space.low + flat_action_space.high)

    def __call__(self, observation, algo_info, testing = False):
        if testing:
            deterministic = True
        elif algo_info is not None and "be_deterministic" in algo_info:
            deterministic = algo_info["be_deterministic"]
        else:
            deterministic = False

        # switches noise off or on if Tianshou policy is DDPG
        if deterministic:
            self._ts_policy.eval()
        else:
            self._ts_policy.train()

        with torch.no_grad():
            # if not deterministic, return fully random action in a fraction of all queries
            if not deterministic and self._fully_random_fraction > 0. and \
                    np.random.rand() < self._fully_random_fraction:
                action = self.action_space.sample()
            else:
                policy_input_obs = self.observation_space.flatten_value(observation)
                batch = Batch(obs = policy_input_obs[None], info={})
                if deterministic:
                    net_output, _ = self._hidden_net(batch.obs)
                    if isinstance(self._hidden_net, SACNetwork):
                        mu, sigma = net_output
                        # map from R^n to action space
                        action = torch.tanh(mu[0])
                    else:
                        action = net_output[0]
                    action = action.detach().numpy()
                else:
                    action = self._ts_policy.forward(batch).act[0].detach().numpy()

                # map action from [-1, 1]^n to flat action space 
                # (NOTE: This has no effect on entropy in SAC)
                action = action*self._action_scale + self._action_bias

                # map flat action back to original action space
                action = self.action_space.unflatten_value(action)

        return 0, action

    def entropy(observation):
        policy_input_obs = self.observation_space.flatten_value(observation)
        batch = Batch(obs = policy_input_obs[None], info={})
        net_output, _ = self._hidden_net(batch.obs)
        mu, sigma = net_output
        return torch.log(sigma*torch.sqrt(2.*torch.pi)).sum() + 0.5*sigma.size()





class EnvInfo:
    def __init__(self, reward = None, new_obs = None, done = None, info = None):
        self.reward = reward
        self.new_obs = new_obs
        self.done = done
        self.info = info



import torch
from torch import nn

class SACNetwork(nn.Module):
    """Outputs mean, variance and state (None) as required by Tianshou implementation of SAC."""
    def __init__(self, input_processing_fn, n_hidden, output_dim, device, 
            clip_range_sigma=None, clip_range_mu=None):
        """
        Args: 
            input_processing_fn: pytorch module which outputs a tensor with shape (batch_size, n_hidden).
        """
        super().__init__()
        self.device = device
        self.input_processing_fn = input_processing_fn
        self.mu_fn = nn.Linear(n_hidden, output_dim)
        if clip_range_sigma is None:
            self.clip_range_sigma = [1.0e-3, 6.0e0]
        else:
            self.clip_range_sigma = clip_range_sigma
        if clip_range_mu is None:
            # Due to squashing this is  close to (-1, 1)
            self.clip_range_mu = [-3., 3.]
        else:
            self.clip_range_mu = clip_range_mu
        self.sigma_fn = nn.Linear(n_hidden, output_dim)

    def forward(self, s, **kwargs):
        if not isinstance(s, torch.Tensor):
            s = torch.tensor(s, device = self.device, dtype = torch.float)
        batch_size = s.shape[0]
        s_flat = s.view(batch_size, -1)
        hidden_output = self.input_processing_fn(s_flat)
        mu = self.mu_fn(hidden_output)
        sigma = self.sigma_fn(hidden_output).exp()
        sigma = torch.clamp(sigma, min=self.clip_range_sigma[0], 
            max=self.clip_range_sigma[1])
        mu = torch.clamp(mu, min=self.clip_range_mu[0], 
            max=self.clip_range_mu[1])
        return (mu, sigma), None



from abc import ABC, abstractmethod

class Model:
    @abstractmethod
    def create(self, state_dim, action_dim):
        pass

    @property
    @abstractmethod
    def actor(self):
        pass

    @property
    @abstractmethod
    def actor_optim(self):
        pass

    @property
    def critics(self):
        return None

    @property
    def critic_optims(self):
        return None

    @abstractmethod
    def get_parameters(self):
        """Returns dictionary containing all learned parameters."""
        pass

    @abstractmethod
    def get_state(self):
        """Returns dictionary containing state."""
        pass

    @abstractmethod
    def load_parameters(self, params):
        """Load all learned parameters from dictionary."""
        pass




import torch
from torch import nn
import torch.nn.functional as F

class MLPNetwork(nn.Module):
    """Multi layer perceptron."""
    def __init__(self, input_dim, units_per_layer, activation_fns, device = "cpu"):
        """
        Args: 
            activation_fns: Either list of or a single activation function from torch.nn. If a list is provided, it has to have the same length as units_per_layer. An entry None corresponds to no activation function.
        """
        super(MLPNetwork, self).__init__()

        self.device = device

        # if only one activation function is given, apply it to all layers
        if type(activation_fns) is not list:
            activation_fns = [activation_fns]*len(units_per_layer)

        assert len(units_per_layer) == len(activation_fns)

        self.layers = []
        in_dim = input_dim
        for n_units, act in zip(units_per_layer, activation_fns):
            self.layers.append(nn.Linear(in_dim, n_units))
            if act is not None:
                self.layers.append(act)
            in_dim = n_units
        self.model = nn.Sequential(*self.layers)

    def forward(self, x, **kwargs):
        if not isinstance(x, torch.Tensor):
            x = torch.tensor(x, device = self.device, dtype = torch.float)
        batch_size = x.shape[0]
        res = x.view(batch_size, -1)
        res = self.model(res)
        return res



from abc import ABC, abstractmethod


class GoalSamplingStrategy(ABC):

    @abstractmethod
    def __call__(self, episode_transitions, n_hindsight_goals):
        """Return a list of indices of transitions from which goals are generated.

        Hindsight goals are constructed from the achieved goal attribute of the 
        selected transitions."""
        pass





import numpy as np

from . import Graph
from ..subtasks import EnvSPSubtaskSpec
from ..nodes import HACNode
from ..utils import listify

class HACGraph(Graph):

    def __init__(self, name, n_layers, env, subtask_specs, HAC_kwargs, update_sgs_rendering=None):
        """
        Args:
        n_layers (int): Number of layers in hierarchy.
        """

        self.n_layers = n_layers
        self._update_sgs_rendering = update_sgs_rendering 

        # convert HAC parameters to list with one entry per layer if not already a list
        HAC_kwargs_list = listify(HAC_kwargs, n_layers)
        subtask_specs = listify(subtask_specs, n_layers - 1)

        # by default penalty for choosing a goal which was not reached by the child is minus
        # maximum number of actions in episode on this layer
        for kwargs, subtask_spec in zip(HAC_kwargs_list, subtask_specs):
            if "child_failure_penalty" not in kwargs:
                kwargs["child_failure_penalty"] = -subtask_spec.max_n_actions

        # make sure testing fraction in lowest level is 0
        HAC_kwargs_list[0]["testing_fraction"] = 0.

        # construct graph
        parents = []
        self._nodes = []
        for l in range(n_layers - 1, -1, -1):
            node_name = "hac_node_layer_{}".format(l)
            node = HACNode(
                    name=node_name, 
                    parents=parents,
                    subtask_spec=subtask_specs[l], 
                    HAC_kwargs=HAC_kwargs_list[l])
            if l < n_layers - 1:
                self._nodes[0].add_child(node)
            self._nodes.insert(0, node)
            parents = [node]

        entry_node = self._nodes[-1]

        super(HACGraph, self).__init__(name, entry_node, env.action_space)

    def get_atomic_action(self, env_obs, sess_info, start_node=None, testing=False):
        action = super().get_atomic_action(env_obs, sess_info, start_node, testing)

        # update subgoals via callback for rendering them in environment
        if len(self._nodes) > 1 and self._update_sgs_rendering is not None:
            self._update_sgs_rendering([nd.current_goal for nd 
                in self._nodes[:-1]])

        return action




from abc import ABC, abstractmethod

from .subtask_specs import SubtaskSpec
from ..spaces import space_from_gym_space


class ReturnMaximSubtaskSpec(SubtaskSpec):
    """Specification of return maximization subtask, i.e., obs, reward etc."""

    def __init__(self, max_n_actions=None):
        super().__init__()
        self._max_n_actions = max_n_actions

    @property
    @abstractmethod
    def obs_space(self):
        """Observation space."""

        pass

    @abstractmethod
    def map_to_obs(self, env_obs, parent_info, ep_step):
        """Maps environment observation, parent info and episode step to observation."""

        pass

    @abstractmethod
    def get_reward(self, obs, act, new_obs, parent_info, ep_step, env_info):
        """Get reward of transition."""

        pass

    @abstractmethod
    def step_update(self, env_info, parent_info, sess_info):
        """Update internal state of subtask spec in each time step."""

        pass

    def reset(self):
        """Reset state of subtask sepc."""

        pass

    @property
    def max_n_actions(self):
        """Maximum number of actions before subtask ends."""

        return self._max_n_actions


class EnvRMSubtaskSpec(ReturnMaximSubtaskSpec):
    """Return maximization subtask specification based on environment."""

    def __init__(self, env, max_n_actions=None):
        super().__init__(max_n_actions)
        self._env = env
        self._env_obs_space = space_from_gym_space(env.observation_space)
        self._cum_reward = 0.

    @property
    def obs_space(self):
        """Use environment observation space."""

        return self._env_obs_space

    def map_to_obs(self, env_obs, parent_info, ep_step):
        """Return environment observation."""

        return env_obs

    def get_reward(self, obs, act, new_obs, parent_info, ep_step, env_info):
        """Return cumulative environment reward.
        
        Reward has been accumulated since las call to get_reward 
        or beginning of episode.
        """

        rew = self._cum_reward + env_info.reward
        self._cum_reward = 0

        return rew

    def step_update(self, env_info, parent_info, sess_info):
        """Add env reward of time step to cumulative reward."""

        self._cum_reward += env_info.reward



from copy import copy

import numpy as np

from . import Subtask


class ReturnMaximSubtask(Subtask):
    """Subtask is to maximize expected return."""

    def __init__(self, name, subtask_spec):
        """Args:
            subtask_spec (ReturnMaximSubtaskSpec): Specification of the return maximization subtask.
        """

        super().__init__(name, subtask_spec)
        self.task_spec = subtask_spec

        self._n_actions_taken = 0
        self._return = 0
        #self._min = np.array([np.inf]*8)
        #self._max = np.array([-np.inf]*8)

    @property
    def observation_space(self):
        return self.task_spec.obs_space

    @property
    def parent_action_space(self):
        return self.task_spec.parent_action_space

    def reset(self):
        self._n_actions_taken = 0
        self._return = 0
        self.task_spec.reset()

    def get_observation(self, env_obs, parent_info, sess_info):
        return self.task_spec.map_to_obs(env_obs, parent_info, sess_info.ep_step)

    def check_interruption(self, env_info, new_subtask_obs, parent_info, sess_info):
        # update internal state of subtask spec in each time step
        self.task_spec.step_update(env_info, parent_info, sess_info)
        # return maximization subtask does not interrupt child nodes
        return False

    def evaluate_transition(self, env_obs, env_info, subtask_trans, parent_info, algo_info, sess_info):
        new_obs = self.task_spec.map_to_obs(
            env_info.new_obs, parent_info, sess_info.ep_step
        )
        #self._min = np.minimum(self._min, new_obs)
        #self._max = np.maximum(self._max, new_obs)
        #print(self._max)
        reward = self.task_spec.get_reward(
            subtask_trans.obs, subtask_trans.action, new_obs, parent_info, 
            sess_info.ep_step, env_info
        )
        # add auxiliary rewards
        if new_obs is not None and subtask_trans.action is not None:
            reward += self.get_aux_rewards(new_obs, subtask_trans.action)
        self._n_actions_taken += 1
        self._return += reward

        # subtask ended if running out of actions
        if (self.task_spec.max_n_actions is not None and 
            self._n_actions_taken >= self.task_spec.max_n_actions):
            # tensorboard logging
            if self._tb_writer is not None:
                    mode = "test" if sess_info.testing else "train"
                    self._tb_writer.add_scalar(f"{self.name}/{mode}/return", self._return, sess_info.total_step)
            self._return = 0
            self._n_actions_taken = 0
            ended = True
        else:
            ended = False

        info = {}
        feedback = {}

        complete_subtask_trans = copy(subtask_trans)
        complete_subtask_trans.reward = reward
        complete_subtask_trans.ended = ended
        complete_subtask_trans.info = info
        

        return complete_subtask_trans, feedback



class SubtaskTransition():
    def __init__(self, obs = None, action = None, ind_child = None, reward = None, 
            new_obs = None, ended = None, info = None):
        self.obs = obs
        self.action = action
        self.ind_child = ind_child
        self.reward = reward
        self.new_obs = new_obs
        self.ended = ended
        self.info = info

    def __str__(self):
        return "obs: {}\n".format(self.obs) + \
        "action: {}\n".format(self.action) + \
        "ind_child: {}\n".format(self.ind_child) + \
        "reward: {}\n".format(self.reward) + \
        "new_obs: {}\n".format(self.new_obs) + \
        "ended: {}\n".format(self.ended) + \
        "info: {}\n".format(self.info)




from gym import Wrapper

from ..spaces import space_from_gym_space


class FlattenWrapper(Wrapper):
    """Flattens observation and action space."""

    def __init__(self, env):
        super().__init__(env)
        self._action_space = space_from_gym_space(env.action_space)
        self._observation_space = space_from_gym_space(env.observation_space)

        self.action_space = self._action_space.get_flat_space().get_gym_space()
        self.observation_space = self._observation_space.get_flat_space().get_gym_space()

    def step(self, action):
        orig_action = self._action_space.unflatten_value(action)
        orig_obs, reward, done, info = self.env.step(orig_action)
        obs = self._observation_space.flatten_value(orig_obs)
        return obs, reward, done, info

    def reset(self, **kwargs):
        orig_obs = self.env.reset(**kwargs)
        obs = self._observation_space.flatten_value(orig_obs)
        return obs






from .subtask import Subtask
from .shortest_path_subtask import ShortestPathSubtask
from. shortest_path_subtask_specs import (ShortestPathSubtaskSpec, BoxSPSubtaskSpec, 
        BoxInfoHidingSPSubtaskSpec, DictSPSubtaskSpec, DictInfoHidingSPSubtaskSpec, EnvSPSubtaskSpec)
from .timed_goal_subtask import TimedGoalSubtask
from .timed_goal_subtask_specs import (TimedGoalSubtaskSpec, BoxTGSubtaskSpec, BoxInfoHidingTGSubtaskSpec, 
        DictInfoHidingTGSubtaskSpec, DictInfoHidingTolTGSubtaskSpec)
from .return_maxim_subtask import ReturnMaximSubtask
from .return_maxim_subtask_specs import ReturnMaximSubtaskSpec, EnvRMSubtaskSpec



class ArrowConfig:
    def __init__(self, scale, width, head_size, color):
        self.scale = scale
        self.width = width
        self.head_size = head_size
        self.color = color

def get_default_subgoal_colors():
    matplotlib_default = ['#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd', 
            '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf']
    rgb_colors = [tuple(int(h[i:i+2], 16)/255. for i in [1, 3, 5]) for h in matplotlib_default]
    return rgb_colors



import numpy as np

from . import AuxReward

class ConstantReward(AuxReward):
    """Constant auxiliary reward."""

    def __init__(self, constant_reward):
        self._reward = constant_reward

    def __call__(self, obs, action, env_reward):
        return self._reward
